#  Copyright 2000-2025 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.


from _pydevd_bundle import pydevd_vars
from _pydevd_bundle.pydevd_constants import NEXT_VALUE_SEPARATOR
from _pydevd_bundle.pydevd_xml import ExceptionOnEvaluate
from _pydevd_bundle.tables.images.pydevd_image_loader import load_image_chunk


class TableCommandType:
    DF_INFO = "DF_INFO"
    SLICE = "SLICE"
    SLICE_CSV = "SLICE_CSV"
    DESCRIBE = "DF_DESCRIBE"
    VISUALIZATION_DATA = "VISUALIZATION_DATA"
    IMAGE_START_CHUNK_LOAD = "IMAGE_START_CHUNK_LOAD"
    IMAGE_CHUNK_LOAD = "IMAGE_CHUNK_LOAD"


def is_error_on_eval(val):
    try:
        # This should be faster than isinstance (but we have to protect against not
        # having a '__class__' attribute).
        is_exception_on_eval = val.__class__ == ExceptionOnEvaluate
    except:
        is_exception_on_eval = False
    return is_exception_on_eval

def exec_image_table_command(init_command, command_type, offset, image_id, f_globals, f_locals):
    table = pydevd_vars.eval_in_context(init_command, f_globals, f_locals)
    is_exception_on_eval = is_error_on_eval(table)
    if is_exception_on_eval:
        return False, table.result

    image_provider = __get_image_provider(table)
    if not image_provider:
        raise RuntimeError('No image provider for: {}'.format(type(table)))

    if command_type == TableCommandType.IMAGE_START_CHUNK_LOAD:
        return True, image_provider.create_image(table)

    return True, load_image_chunk(offset, image_id)


def exec_table_command(init_command, command_type, start_index, end_index, format, f_globals,
                       f_locals):
    table = pydevd_vars.eval_in_context(init_command, f_globals, f_locals)
    is_exception_on_eval = is_error_on_eval(table)
    if is_exception_on_eval:
        return False, table.result

    table_provider = __get_table_provider(table)
    if not table_provider:
        raise RuntimeError('No table data provider for: {}'.format(type(table)))

    res = []
    if command_type == TableCommandType.DF_INFO:
        res.append(table_provider.get_type(table))
        res.append(NEXT_VALUE_SEPARATOR)
        res.append(table_provider.get_shape(table))
        res.append(NEXT_VALUE_SEPARATOR)
        res.append(table_provider.get_head(table))
        res.append(NEXT_VALUE_SEPARATOR)
        res.append(table_provider.get_column_types(table))

    elif command_type == TableCommandType.DESCRIBE:
        res.append(table_provider.get_column_descriptions(table))

    elif command_type == TableCommandType.VISUALIZATION_DATA:
        res.append(table_provider.get_value_occurrences_count(table))
        res.append(NEXT_VALUE_SEPARATOR)

    elif command_type == TableCommandType.SLICE:
        res.append(table_provider.get_data(table, False, start_index, end_index, format))
    elif command_type == TableCommandType.SLICE_CSV:
        res.append(table_provider.get_data(table, True, start_index, end_index, format))

    return True, ''.join(res)


def __get_type_name(table):
    table_data_type = type(table)
    table_data_type_name = '{}.{}'.format(table_data_type.__module__, table_data_type.__name__)
    return table_data_type_name


# noinspection PyUnresolvedReferences
def __get_table_provider(output):
    # type: (str) -> Any
    type_qualified_name = __get_type_name(output)
    numpy_based_type_qualified_names = ['tensorflow.python.framework.ops.EagerTensor',
                                        'tensorflow.python.ops.resource_variable_ops.ResourceVariable',
                                        'tensorflow.python.framework.sparse_tensor.SparseTensor',
                                        'torch.Tensor']
    table_provider = None
    if type_qualified_name in ['pandas.core.frame.DataFrame',
                               'pandas.core.series.Series',
                               'geopandas.geoseries.GeoSeries',
                               'geopandas.geodataframe.GeoDataFrame',
                               'pandera.typing.pandas.DataFrame']:
        import _pydevd_bundle.tables.pydevd_pandas as table_provider
    # dict is needed for sort commands
    elif type_qualified_name == 'builtins.dict':
        table_type_name = __get_type_name(output['data'])
        if table_type_name in numpy_based_type_qualified_names:
            import _pydevd_bundle.tables.pydevd_numpy_based as table_provider
        else:
            import _pydevd_bundle.tables.pydevd_numpy as table_provider
    elif type_qualified_name == 'numpy.ndarray' or type_qualified_name == 'numpy.rec.recarray':
        import _pydevd_bundle.tables.pydevd_numpy as table_provider
    elif type_qualified_name in numpy_based_type_qualified_names:
        import _pydevd_bundle.tables.pydevd_numpy_based as table_provider
    elif type_qualified_name.startswith('polars') and (
            type_qualified_name.endswith('DataFrame')
            or type_qualified_name.endswith('Series')):
        import _pydevd_bundle.tables.pydevd_polars as table_provider
    elif type_qualified_name == 'datasets.arrow_dataset.Dataset':
        import _pydevd_bundle.tables.pydevd_dataset as table_provider

    return table_provider


# noinspection PyUnresolvedReferences
def __get_image_provider(output):
    # type: (str) -> Any
    type_qualified_name = __get_type_name(output)
    numpy_based_type_qualified_names = ['tensorflow.python.framework.ops.EagerTensor',
                                        'tensorflow.python.ops.resource_variable_ops.ResourceVariable',
                                        'tensorflow.python.framework.sparse_tensor.SparseTensor',
                                        'torch.Tensor']
    image_provider = None
    if type_qualified_name == 'builtins.dict':
        table_type_name = __get_type_name(output['data'])
        if table_type_name in numpy_based_type_qualified_names:
            import _pydevd_bundle.tables.images.pydevd_numpy_based_image as image_provider
        else:
            import _pydevd_bundle.tables.images.pydevd_numpy_image as image_provider
    elif type_qualified_name in numpy_based_type_qualified_names:
        import _pydevd_bundle.tables.images.pydevd_numpy_based_image as image_provider
    elif type_qualified_name == 'numpy.ndarray':
        import _pydevd_bundle.tables.images.pydevd_numpy_image as image_provider
    elif type_qualified_name in ['PIL.Image.Image', 'PIL.PngImagePlugin.PngImageFile', 'PIL.JpegImagePlugin.JpegImageFile']:
        import _pydevd_bundle.tables.images.pydevd_pillow_image as image_provider
    elif type_qualified_name in ['matplotlib.figure.Figure', 'plotly.graph_objs._figure.Figure']:
        import _pydevd_bundle.tables.images.pydevd_matplotlib_image as image_provider

    return image_provider



# Defines which version of the trace_dispatch we'll use.
# Should give warning only here if cython is not available but supported.

import os
import sys
import traceback
from _pydevd_bundle.pydevd_constants import CYTHON_SUPPORTED, IS_PYCHARM


use_cython = os.getenv('PYDEVD_USE_CYTHON', None)
dirname = os.path.dirname(os.path.dirname(__file__))
# Do not show incorrect warning for .egg files for Remote debugger
if not CYTHON_SUPPORTED or dirname.endswith('.egg'):
    # Do not try to import cython extensions if cython isn't supported
    use_cython = 'NO'


def delete_old_compiled_extensions():
    import _pydevd_bundle_ext
    cython_extensions_dir = os.path.dirname(os.path.dirname(_pydevd_bundle_ext.__file__))
    _pydevd_bundle_ext_dir = os.path.dirname(_pydevd_bundle_ext.__file__)
    _pydevd_frame_eval_ext_dir = os.path.join(cython_extensions_dir, '_pydevd_frame_eval_ext')
    try:
        import shutil
        for file in os.listdir(_pydevd_bundle_ext_dir):
            if file.startswith("pydevd") and file.endswith(".so"):
                os.remove(os.path.join(_pydevd_bundle_ext_dir, file))
        for file in os.listdir(_pydevd_frame_eval_ext_dir):
            if file.startswith("pydevd") and file.endswith(".so"):
                os.remove(os.path.join(_pydevd_frame_eval_ext_dir, file))
        build_dir = os.path.join(cython_extensions_dir, "build")
        if os.path.exists(build_dir):
            shutil.rmtree(os.path.join(cython_extensions_dir, "build"))
    except OSError:
        from _pydev_bundle.pydev_monkey import log_error_once
        log_error_once("warning: failed to delete old cython speedups. Please delete all *.so files from the directories "
                       "\"%s\" and \"%s\"" % (_pydevd_bundle_ext_dir, _pydevd_frame_eval_ext_dir))

show_tracing_warning = False

if use_cython == 'YES':
    # We must import the cython version if forcing cython
    # noinspection PyUnresolvedReferences
    from _pydevd_bundle.pydevd_cython_wrapper import (
        trace_dispatch as _trace_dispatch,
        global_cache_skips,
        global_cache_frame_skips,
        fix_top_level_trace_and_get_trace_func,
        set_additional_thread_info,
        handle_breakpoint_condition,
        handle_breakpoint_expression, DEBUG_START, DEBUG_START_PY3K,
        should_stop_on_exception, handle_exception, manage_return_values,
    )
    def trace_dispatch(py_db, frame, event, arg):
        if _trace_dispatch is None:
            return None
        return _trace_dispatch(py_db, frame, event, arg)

elif use_cython == 'NO':
    # Use the regular version if not forcing cython
    # noinspection PyUnresolvedReferences
    from _pydevd_bundle.pydevd_trace_dispatch_regular import (
        trace_dispatch,
        global_cache_skips,
        global_cache_frame_skips,
        fix_top_level_trace_and_get_trace_func,
        set_additional_thread_info,
    )
    # noinspection PyUnresolvedReferences
    from _pydevd_bundle.pydevd_frame import (
        handle_breakpoint_condition,
        handle_breakpoint_expression, DEBUG_START, DEBUG_START_PY3K,
        should_stop_on_exception, handle_exception, manage_return_values,
    )

elif use_cython is None:
    # Regular: use fallback if not found and give message to user
    try:
        from _pydevd_bundle.pydevd_cython_wrapper import (
            trace_dispatch as _trace_dispatch,
            global_cache_skips,
            global_cache_frame_skips,
            fix_top_level_trace_and_get_trace_func,
            set_additional_thread_info,
            handle_breakpoint_condition,
            handle_breakpoint_expression, DEBUG_START, DEBUG_START_PY3K,
            should_stop_on_exception, handle_exception, manage_return_values,
        )
        def trace_dispatch(py_db, frame, event, arg):
            if _trace_dispatch is None:
                return None
            return _trace_dispatch(py_db, frame, event, arg)

    except ImportError as e:
        try:
            if hasattr(e, 'version_mismatch'):
                delete_old_compiled_extensions()
            from _pydevd_bundle.pydevd_trace_dispatch_regular import (
                trace_dispatch,
                global_cache_skips,
                global_cache_frame_skips,
                fix_top_level_trace_and_get_trace_func,
                set_additional_thread_info,
            )
            from _pydevd_bundle.pydevd_frame import (
                handle_breakpoint_condition,
                handle_breakpoint_expression, DEBUG_START, DEBUG_START_PY3K,
                should_stop_on_exception, handle_exception, manage_return_values,
            )
            from _pydev_bundle.pydev_monkey import log_error_once

            if not IS_PYCHARM:
                log_error_once("warning: Debugger speedups using cython not found. Run '\"%s\" \"%s\" build_ext --inplace' to build." % (
                    sys.executable, os.path.join(dirname, 'setup_cython.py')))
            else:
                # Check if `setuptools` are available. Without them, it is impossible
                # to build the extensions anyway.
                are_setuptools_available = False
                try:
                    import setuptools
                    are_setuptools_available = True
                except ImportError:
                    pass
                show_tracing_warning = are_setuptools_available
        except Exception as e:
            from _pydev_bundle.pydev_monkey import log_debug
            os.environ['PYDEVD_USE_CYTHON'] = 'NO'
            log_debug("Problems with cython extensions. Debugger speedups using cython are turned off")
            log_debug(traceback.format_exc())
            from _pydevd_bundle.pydevd_additional_thread_info import set_additional_thread_info

else:
    raise RuntimeError('Unexpected value for PYDEVD_USE_CYTHON: %s (accepted: YES, NO)' % (use_cython,))



"""
The UserModuleDeleter and runfile methods are copied from
Spyder and carry their own license agreement.
http://code.google.com/p/spyderlib/source/browse/spyderlib/widgets/externalshell/sitecustomize.py

Spyder License Agreement (MIT License)
--------------------------------------

Copyright (c) 2009-2012 Pierre Raybaut

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
"""

import os
import sys

from _pydev_bundle import pydev_imports
from _pydevd_bundle.pydevd_utils import save_main_module
from pydevd_file_utils import get_fullname, rPath


# The following classes and functions are mainly intended to be used from
# an interactive Python session
class UserModuleDeleter:
    """
    User Module Deleter (UMD) aims at deleting user modules
    to force Python to deeply reload them during import

    pathlist [list]: blacklist in terms of module path
    namelist [list]: blacklist in terms of module name
    """
    def __init__(self, namelist=None, pathlist=None):
        if namelist is None:
            namelist = []
        self.namelist = namelist
        if pathlist is None:
            pathlist = []
        self.pathlist = pathlist
        try:
            # blacklist all files in org.python.pydev/pysrc
            import pydev_pysrc, inspect
            self.pathlist.append(os.path.dirname(pydev_pysrc.__file__))
        except:
            pass
        self.previous_modules = list(sys.modules.keys())

    def is_module_blacklisted(self, modname, modpath):
        for path in [sys.prefix] + self.pathlist:
            if modpath.startswith(path):
                return True
        else:
            return set(modname.split('.')) & set(self.namelist)

    def run(self, verbose=False):
        """
        Del user modules to force Python to deeply reload them

        Do not del modules which are considered as system modules, i.e.
        modules installed in subdirectories of Python interpreter's binary
        Do not del C modules
        """
        log = []
        modules_copy = dict(sys.modules)
        for modname, module in modules_copy.items():
            if modname == 'aaaaa':
                print(modname, module)
                print(self.previous_modules)
            if modname not in self.previous_modules:
                modpath = getattr(module, '__file__', None)
                if modpath is None:
                    # *module* is a C module that is statically linked into the
                    # interpreter. There is no way to know its path, so we
                    # choose to ignore it.
                    continue
                if not self.is_module_blacklisted(modname, modpath):
                    log.append(modname)
                    del sys.modules[modname]
        if verbose and log:
            print("\x1b[4;33m%s\x1b[24m%s\x1b[0m" % ("UMD has deleted",
                                                     ": " + ", ".join(log)))

__umd__ = None

_get_globals_callback = None


def _set_globals_function(get_globals):
    global _get_globals_callback
    _get_globals_callback = get_globals


def _get_interpreter_globals():
    """Return current Python interpreter globals namespace"""
    if _get_globals_callback is not None:
        return _get_globals_callback()
    else:
        try:
            from __main__ import __dict__ as namespace
        except ImportError:
            try:
                # The import fails on IronPython
                import __main__
                namespace = __main__.__dict__
            except:
                namespace
        shell = namespace.get('__ipythonshell__')
        if shell is not None and hasattr(shell, 'user_ns'):
            # IPython 0.12+ kernel
            return shell.user_ns
        else:
            # Python interpreter
            return namespace


def runfile(filename, args=None, wdir=None, is_module=False, global_vars=None):
    """
    Run filename
    args: command line arguments (string)
    wdir: working directory
    """
    try:
        if hasattr(filename, 'decode'):
            filename = filename.decode('utf-8')
    except (UnicodeError, TypeError):
        pass

    global __umd__
    if os.environ.get("PYDEV_UMD_ENABLED", "").lower() == "true":
        if __umd__ is None:
            namelist = os.environ.get("PYDEV_UMD_NAMELIST", None)
            if namelist is not None:
                namelist = namelist.split(',')
            __umd__ = UserModuleDeleter(namelist=namelist)
        else:
            verbose = os.environ.get("PYDEV_UMD_VERBOSE", "").lower() == "true"
            __umd__.run(verbose=verbose)

    if global_vars is None:
        m = save_main_module(filename, 'pydev_umd')
        global_vars = m.__dict__
        try:
            global_vars['__builtins__'] = __builtins__
        except NameError:
            pass  # Not there on Jython...

    local_vars = global_vars

    module_name = None
    entry_point_fn = None
    if is_module:
        file, _, entry_point_fn = filename.partition(':')
        module_name = file
        filename = get_fullname(file)
        if filename is None:
            sys.stderr.write("No module named %s\n" % file)
            return
    else:
        # The file being run must be in the pythonpath (even if it was not before)
        sys.path.insert(0, os.path.split(rPath(filename))[0])

    global_vars['__file__'] = filename

    sys.argv = [filename]
    if args is not None:
        for arg in args:
            sys.argv.append(arg)

    if wdir is not None:
        try:
            if hasattr(wdir, 'decode'):
                wdir = wdir.decode('utf-8')
        except (UnicodeError, TypeError):
            pass
        os.chdir(wdir)

    try:
        if not is_module:
            pydev_imports.execfile(filename, global_vars, local_vars)  # execute the script
        else:
            # treat ':' as a seperator between module and entry point function
            # if there is no entry point we run we same as with -m switch. Otherwise we perform
            # an import and execute the entry point
            if entry_point_fn:
                mod = __import__(module_name, level=0, fromlist=[entry_point_fn], globals=global_vars, locals=local_vars)
                func = getattr(mod, entry_point_fn)
                func()
            else:
                # Run with the -m switch
                import runpy
                if hasattr(runpy, '_run_module_as_main'):
                    runpy._run_module_as_main(module_name)
                else:
                    runpy.run_module(module_name)
    finally:
        sys.argv = ['']
        interpreter_globals = _get_interpreter_globals()
        interpreter_globals.update(global_vars)



#  Copyright 2000-2024 JetBrains s.r.o. and contributors. Use of this source code is governed by the Apache 2.0 license.
"""Bytecode analyzing utils for extracting smart steps into variants."""

from collections import namedtuple

from _pydevd_bundle.pydevd_constants import IS_PY311_OR_GREATER, IS_PY36_OR_GREATER, \
    IS_PY2, IS_CPYTHON

__all__ = [
    'find_stepping_variants',
    'get_stepping_variants',
]


def _get_stepping_variants_dummy(_):
    return []


get_stepping_variants = _get_stepping_variants_dummy

if not IS_CPYTHON:
    pass
elif IS_PY2:
    from _pydevd_bundle.smart_step_into.pydevd_smart_step_into_27 import (
        get_stepping_variants)
else:
    from _pydevd_bundle.smart_step_into.pydevd_smart_step_into_util import get_stepping_variants

_Variant = namedtuple('Variant', ['name', 'is_visited'])


def find_stepping_variants(frame, start_line, end_line):
    """Finds the *ordered* stepping targets within the given line range."""
    code = frame.f_code
    last_instruction = frame.f_lasti
    for instruction in get_stepping_variants(code):
        if not instruction.lineno or instruction.lineno > end_line:
            continue
        if start_line <= instruction.lineno <= end_line:
            yield _Variant(instruction.argval, instruction.offset <= last_instruction)



#!~/.wine/drive_c/Python25/python.exe
# -*- coding: utf-8 -*-

# Copyright (c) 2009-2014, Mario Vilas
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice,this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
Registry access.

@group Instrumentation:
    Registry, RegistryKey
"""

from __future__ import with_statement

__revision__ = "$Id$"

__all__ = ['Registry']

import sys
from winappdbg import win32
from winappdbg import compat
import collections
import warnings

#==============================================================================

class _RegistryContainer (object):
    """
    Base class for L{Registry} and L{RegistryKey}.
    """

    # Dummy object to detect empty arguments.
    class __EmptyArgument:
        pass
    __emptyArgument = __EmptyArgument()

    def __init__(self):
        self.__default = None

    def has_key(self, name):
        return name in self

    def get(self, name, default=__emptyArgument):
        try:
            return self[name]
        except KeyError:
            if default is RegistryKey.__emptyArgument:
                return self.__default
            return default

    def setdefault(self, default):
        self.__default = default

    def __iter__(self):
        return compat.iterkeys(self)

#==============================================================================

class RegistryKey (_RegistryContainer):
    """
    Exposes a single Windows Registry key as a dictionary-like object.

    @see: L{Registry}

    @type path: str
    @ivar path: Registry key path.

    @type handle: L{win32.RegistryKeyHandle}
    @ivar handle: Registry key handle.
    """

    def __init__(self, path, handle):
        """
        @type  path: str
        @param path: Registry key path.

        @type  handle: L{win32.RegistryKeyHandle}
        @param handle: Registry key handle.
        """
        super(RegistryKey, self).__init__()
        if path.endswith('\\'):
            path = path[:-1]
        self._path   = path
        self._handle = handle

    @property
    def path(self):
        return self._path

    @property
    def handle(self):
        #if not self._handle:
        #    msg = "This Registry key handle has already been closed."
        #    raise RuntimeError(msg)
        return self._handle

    #def close(self):
    #    """
    #    Close the Registry key handle, freeing its resources. It cannot be
    #    used again after calling this method.
    #
    #    @note: This method will be called automatically by the garbage
    #        collector, and upon exiting a "with" block.
    #
    #    @raise RuntimeError: This Registry key handle has already been closed.
    #    """
    #    self.handle.close()
    #
    #def __enter__(self):
    #    """
    #    Compatibility with the "C{with}" Python statement.
    #    """
    #    return self
    #
    #def __exit__(self, type, value, traceback):
    #    """
    #    Compatibility with the "C{with}" Python statement.
    #    """
    #    try:
    #        self.close()
    #    except Exception:
    #        pass

    def __contains__(self, name):
        try:
            win32.RegQueryValueEx(self.handle, name, False)
            return True
        except WindowsError:
            e = sys.exc_info()[1]
            if e.winerror == win32.ERROR_FILE_NOT_FOUND:
                return False
            raise

    def __getitem__(self, name):
        try:
            return win32.RegQueryValueEx(self.handle, name)[0]
        except WindowsError:
            e = sys.exc_info()[1]
            if e.winerror == win32.ERROR_FILE_NOT_FOUND:
                raise KeyError(name)
            raise

    def __setitem__(self, name, value):
        win32.RegSetValueEx(self.handle, name, value)

    def __delitem__(self, name):
        win32.RegDeleteValue(self.handle, name)

    def iterkeys(self):
        handle = self.handle
        index = 0
        while 1:
            resp = win32.RegEnumValue(handle, index, False)
            if resp is None:
                break
            yield resp[0]
            index += 1

    def itervalues(self):
        handle = self.handle
        index = 0
        while 1:
            resp = win32.RegEnumValue(handle, index)
            if resp is None:
                break
            yield resp[2]
            index += 1

    def iteritems(self):
        handle = self.handle
        index = 0
        while 1:
            resp = win32.RegEnumValue(handle, index)
            if resp is None:
                break
            yield resp[0], resp[2]
            index += 1

    def keys(self):
        # return list(self.iterkeys())   # that can't be optimized by psyco
        handle = self.handle
        keys = list()
        index = 0
        while 1:
            resp = win32.RegEnumValue(handle, index, False)
            if resp is None:
                break
            keys.append(resp[0])
            index += 1
        return keys

    def values(self):
        # return list(self.itervalues()) # that can't be optimized by psyco
        handle = self.handle
        values = list()
        index = 0
        while 1:
            resp = win32.RegEnumValue(handle, index)
            if resp is None:
                break
            values.append(resp[2])
            index += 1
        return values

    def items(self):
        # return list(self.iteritems()) # that can't be optimized by psyco
        handle = self.handle
        items = list()
        index = 0
        while 1:
            resp = win32.RegEnumValue(handle, index)
            if resp is None:
                break
            items.append( (resp[0], resp[2]) )
            index += 1
        return items

    def get_value_type(self, name):
        """
        Retrieves the low-level data type for the given value.

        @type  name: str
        @param name: Registry value name.

        @rtype:  int
        @return: One of the following constants:
         - L{win32.REG_NONE} (0)
         - L{win32.REG_SZ} (1)
         - L{win32.REG_EXPAND_SZ} (2)
         - L{win32.REG_BINARY} (3)
         - L{win32.REG_DWORD} (4)
         - L{win32.REG_DWORD_BIG_ENDIAN} (5)
         - L{win32.REG_LINK} (6)
         - L{win32.REG_MULTI_SZ} (7)
         - L{win32.REG_RESOURCE_LIST} (8)
         - L{win32.REG_FULL_RESOURCE_DESCRIPTOR} (9)
         - L{win32.REG_RESOURCE_REQUIREMENTS_LIST} (10)
         - L{win32.REG_QWORD} (11)

        @raise KeyError: The specified value could not be found.
        """
        try:
            return win32.RegQueryValueEx(self.handle, name)[1]
        except WindowsError:
            e = sys.exc_info()[1]
            if e.winerror == win32.ERROR_FILE_NOT_FOUND:
                raise KeyError(name)
            raise

    def clear(self):
        handle = self.handle
        while 1:
            resp = win32.RegEnumValue(handle, 0, False)
            if resp is None:
                break
            win32.RegDeleteValue(handle, resp[0])

    def __str__(self):
        default = self['']
        return str(default)

    def __unicode__(self):
        default = self[u'']
        return compat.unicode(default)

    def __repr__(self):
        return '<Registry key: "%s">' % self._path

    def iterchildren(self):
        """
        Iterates the subkeys for this Registry key.

        @rtype:  iter of L{RegistryKey}
        @return: Iterator of subkeys.
        """
        handle = self.handle
        index = 0
        while 1:
            subkey = win32.RegEnumKey(handle, index)
            if subkey is None:
                break
            yield self.child(subkey)
            index += 1

    def children(self):
        """
        Returns a list of subkeys for this Registry key.

        @rtype:  list(L{RegistryKey})
        @return: List of subkeys.
        """
        # return list(self.iterchildren()) # that can't be optimized by psyco
        handle = self.handle
        result = []
        index = 0
        while 1:
            subkey = win32.RegEnumKey(handle, index)
            if subkey is None:
                break
            result.append( self.child(subkey) )
            index += 1
        return result

    def child(self, subkey):
        """
        Retrieves a subkey for this Registry key, given its name.

        @type  subkey: str
        @param subkey: Name of the subkey.

        @rtype:  L{RegistryKey}
        @return: Subkey.
        """
        path = self._path + '\\' + subkey
        handle = win32.RegOpenKey(self.handle, subkey)
        return RegistryKey(path, handle)

    def flush(self):
        """
        Flushes changes immediately to disk.

        This method is normally not needed, as the Registry writes changes
        to disk by itself. This mechanism is provided to ensure the write
        happens immediately, as opposed to whenever the OS wants to.

        @warn: Calling this method too often may degrade performance.
        """
        win32.RegFlushKey(self.handle)

#==============================================================================

# TODO: possibly cache the RegistryKey objects
# to avoid opening and closing handles many times on code sequences like this:
#
# r = Registry()
# r['HKLM\\Software\\Microsoft\\Windows NT\\CurrentVersion\\Run']['Example 1'] = 'example1.exe'
# r['HKLM\\Software\\Microsoft\\Windows NT\\CurrentVersion\\Run']['Example 2'] = 'example2.exe'
# r['HKLM\\Software\\Microsoft\\Windows NT\\CurrentVersion\\Run']['Example 3'] = 'example3.exe'

# TODO: support for access flags?
# TODO: should be possible to disable the safety checks (see __delitem__)

# TODO: workaround for an API bug described by a user in MSDN
#
# http://msdn.microsoft.com/en-us/library/windows/desktop/aa379776(v=vs.85).aspx
#
# Apparently RegDeleteTree won't work remotely from Win7 to WinXP, and the only
# solution is to recursively call RegDeleteKey.

class Registry (_RegistryContainer):
    """
    Exposes the Windows Registry as a Python container.

    @type machine: str or None
    @ivar machine: For a remote Registry, the machine name.
        For a local Registry, the value is C{None}.
    """

    _hives_by_name = {

        # Short names
        'HKCR' : win32.HKEY_CLASSES_ROOT,
        'HKCU' : win32.HKEY_CURRENT_USER,
        'HKLM' : win32.HKEY_LOCAL_MACHINE,
        'HKU'  : win32.HKEY_USERS,
        'HKPD' : win32.HKEY_PERFORMANCE_DATA,
        'HKCC' : win32.HKEY_CURRENT_CONFIG,

        # Long names
        'HKEY_CLASSES_ROOT'     : win32.HKEY_CLASSES_ROOT,
        'HKEY_CURRENT_USER'     : win32.HKEY_CURRENT_USER,
        'HKEY_LOCAL_MACHINE'    : win32.HKEY_LOCAL_MACHINE,
        'HKEY_USERS'            : win32.HKEY_USERS,
        'HKEY_PERFORMANCE_DATA' : win32.HKEY_PERFORMANCE_DATA,
        'HKEY_CURRENT_CONFIG'   : win32.HKEY_CURRENT_CONFIG,
    }

    _hives_by_value = {
        win32.HKEY_CLASSES_ROOT     : 'HKEY_CLASSES_ROOT',
        win32.HKEY_CURRENT_USER     : 'HKEY_CURRENT_USER',
        win32.HKEY_LOCAL_MACHINE    : 'HKEY_LOCAL_MACHINE',
        win32.HKEY_USERS            : 'HKEY_USERS',
        win32.HKEY_PERFORMANCE_DATA : 'HKEY_PERFORMANCE_DATA',
        win32.HKEY_CURRENT_CONFIG   : 'HKEY_CURRENT_CONFIG',
    }

    _hives = sorted(compat.itervalues(_hives_by_value))

    def __init__(self, machine = None):
        """
        Opens a local or remote registry.

        @type  machine: str
        @param machine: Optional machine name. If C{None} it opens the local
            registry.
        """
        self._machine = machine
        self._remote_hives = {}

    @property
    def machine(self):
        return self._machine

    def _split_path(self, path):
        """
        Splits a Registry path and returns the hive and key.

        @type  path: str
        @param path: Registry path.

        @rtype:  tuple( int, str )
        @return: Tuple containing the hive handle and the subkey path.
            The hive handle is always one of the following integer constants:
             - L{win32.HKEY_CLASSES_ROOT}
             - L{win32.HKEY_CURRENT_USER}
             - L{win32.HKEY_LOCAL_MACHINE}
             - L{win32.HKEY_USERS}
             - L{win32.HKEY_PERFORMANCE_DATA}
             - L{win32.HKEY_CURRENT_CONFIG}
        """
        if '\\' in path:
            p = path.find('\\')
            hive = path[:p]
            path = path[p+1:]
        else:
            hive = path
            path = None
        handle = self._hives_by_name[ hive.upper() ]
        return handle, path

    def _parse_path(self, path):
        """
        Parses a Registry path and returns the hive and key.

        @type  path: str
        @param path: Registry path.

        @rtype:  tuple( int, str )
        @return: Tuple containing the hive handle and the subkey path.
            For a local Registry, the hive handle is an integer.
            For a remote Registry, the hive handle is a L{RegistryKeyHandle}.
        """
        handle, path = self._split_path(path)
        if self._machine is not None:
            handle = self._connect_hive(handle)
        return handle, path

    def _join_path(self, hive, subkey):
        """
        Joins the hive and key to make a Registry path.

        @type  hive: int
        @param hive: Registry hive handle.
            The hive handle must be one of the following integer constants:
             - L{win32.HKEY_CLASSES_ROOT}
             - L{win32.HKEY_CURRENT_USER}
             - L{win32.HKEY_LOCAL_MACHINE}
             - L{win32.HKEY_USERS}
             - L{win32.HKEY_PERFORMANCE_DATA}
             - L{win32.HKEY_CURRENT_CONFIG}

        @type  subkey: str
        @param subkey: Subkey path.

        @rtype:  str
        @return: Registry path.
        """
        path = self._hives_by_value[hive]
        if subkey:
            path = path + '\\' + subkey
        return path

    def _sanitize_path(self, path):
        """
        Sanitizes the given Registry path.

        @type  path: str
        @param path: Registry path.

        @rtype:  str
        @return: Registry path.
        """
        return self._join_path( *self._split_path(path) )

    def _connect_hive(self, hive):
        """
        Connect to the specified hive of a remote Registry.

        @note: The connection will be cached, to close all connections and
            erase this cache call the L{close} method.

        @type  hive: int
        @param hive: Hive to connect to.

        @rtype:  L{win32.RegistryKeyHandle}
        @return: Open handle to the remote Registry hive.
        """
        try:
            handle = self._remote_hives[hive]
        except KeyError:
            handle = win32.RegConnectRegistry(self._machine, hive)
            self._remote_hives[hive] = handle
        return handle

    def close(self):
        """
        Closes all open connections to the remote Registry.

        No exceptions are raised, even if an error occurs.

        This method has no effect when opening the local Registry.

        The remote Registry will still be accessible after calling this method
        (new connections will be opened automatically on access).
        """
        while self._remote_hives:
            hive = self._remote_hives.popitem()[1]
            try:
                hive.close()
            except Exception:
                try:
                    e = sys.exc_info()[1]
                    msg = "Cannot close registry hive handle %s, reason: %s"
                    msg %= (hive.value, str(e))
                    warnings.warn(msg)
                except Exception:
                    pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __repr__(self):
        if self._machine:
            return '<Remote Registry at "%s">' % self._machine
        return '<Local Registry>'

    def __contains__(self, path):
        hive, subpath = self._parse_path(path)
        try:
            with win32.RegOpenKey(hive, subpath):
                return True
        except WindowsError:
            e = sys.exc_info()[1]
            if e.winerror == win32.ERROR_FILE_NOT_FOUND:
                return False
            raise

    def __getitem__(self, path):
        path = self._sanitize_path(path)
        hive, subpath = self._parse_path(path)
        try:
            handle = win32.RegOpenKey(hive, subpath)
        except WindowsError:
            e = sys.exc_info()[1]
            if e.winerror == win32.ERROR_FILE_NOT_FOUND:
                raise KeyError(path)
            raise
        return RegistryKey(path, handle)

    def __setitem__(self, path, value):
        do_copy = isinstance(value, RegistryKey)
        if not do_copy and not isinstance(value, str)    \
                       and not isinstance(value, compat.unicode):
            if isinstance(value, object):
                t = value.__class__.__name__
            else:
                t = type(value)
            raise TypeError("Expected string or RegistryKey, got %s" % t)
        hive, subpath = self._parse_path(path)
        with win32.RegCreateKey(hive, subpath) as handle:
            if do_copy:
                win32.RegCopyTree(value.handle, None, handle)
            else:
                win32.RegSetValueEx(handle, None, value)

    # XXX FIXME currently not working!
    # It's probably best to call RegDeleteKey recursively, even if slower.
    def __delitem__(self, path):
        hive, subpath = self._parse_path(path)
        if not subpath:
            raise TypeError(
                "Are you SURE you want to wipe out an entire hive?!"
                " Call win32.RegDeleteTree() directly if you must...")
        try:
            win32.RegDeleteTree(hive, subpath)
        except WindowsError:
            e = sys.exc_info()[1]
            if e.winerror == win32.ERROR_FILE_NOT_FOUND:
                raise KeyError(path)
            raise

    def create(self, path):
        """
        Creates a new Registry key.

        @type  path: str
        @param path: Registry key path.

        @rtype:  L{RegistryKey}
        @return: The newly created Registry key.
        """
        path = self._sanitize_path(path)
        hive, subpath = self._parse_path(path)
        handle = win32.RegCreateKey(hive, subpath)
        return RegistryKey(path, handle)

    def subkeys(self, path):
        """
        Returns a list of subkeys for the given Registry key.

        @type  path: str
        @param path: Registry key path.

        @rtype:  list(str)
        @return: List of subkey names.
        """
        result = list()
        hive, subpath = self._parse_path(path)
        with win32.RegOpenKey(hive, subpath) as handle:
            index = 0
            while 1:
                name = win32.RegEnumKey(handle, index)
                if name is None:
                    break
                result.append(name)
                index += 1
        return result

    def iterate(self, path):
        """
        Returns a recursive iterator on the specified key and its subkeys.

        @type  path: str
        @param path: Registry key path.

        @rtype:  iterator
        @return: Recursive iterator that returns Registry key paths.

        @raise KeyError: The specified path does not exist.
        """
        if path.endswith('\\'):
            path = path[:-1]
        if not self.has_key(path):
            raise KeyError(path)
        stack = collections.deque()
        stack.appendleft(path)
        return self.__iterate(stack)

    def iterkeys(self):
        """
        Returns an iterator that crawls the entire Windows Registry.
        """
        stack = collections.deque(self._hives)
        stack.reverse()
        return self.__iterate(stack)

    def __iterate(self, stack):
        while stack:
            path = stack.popleft()
            yield path
            try:
                subkeys = self.subkeys(path)
            except WindowsError:
                continue
            prefix = path + '\\'
            subkeys = [prefix + name for name in subkeys]
            stack.extendleft(subkeys)



"""Contains methods for building XML structures for interacting with IDE

The methods from this file are used for the debugger interaction. Please note
that Python console now uses Thrift structures with the similar methods
contained in `pydevd_thrift.py` file.
"""
import sys
import traceback

from _pydev_bundle import pydev_log
from _pydev_bundle.pydev_imports import quote
from _pydevd_bundle import pydevd_extension_utils
from _pydevd_bundle import pydevd_resolver
from _pydevd_bundle.pydevd_constants import dict_keys, IS_PY3K, \
    LOAD_VALUES_POLICY, DEFAULT_VALUES_DICT
from _pydevd_bundle.pydevd_extension_api import TypeResolveProvider, \
    StrPresentationProvider
from _pydevd_bundle.pydevd_frame_type_handler import get_vars_handler, \
    DO_NOT_PROCESS_VARS, XML_COMMUNICATION_VARS_HANDLER
from _pydevd_bundle.pydevd_repr_utils import get_value_repr
from _pydevd_bundle.pydevd_user_type_renderers_utils import \
    try_get_type_renderer_for_var
from _pydevd_bundle.pydevd_utils import is_string, should_evaluate_full_value, \
    should_evaluate_shape, is_safe_to_access, is_container_with_shape_dtype

try:
    import types

    frame_type = types.FrameType
except:
    frame_type = None


def make_valid_xml_value(s):
    # Same thing as xml.sax.saxutils.escape but also escaping double quotes.
    return s.replace("&", "&amp;").replace('<', '&lt;').replace('>', '&gt;').replace('"', '&quot;')


class ExceptionOnEvaluate:
    def __init__(self, result):
        self.result = result


_IS_JYTHON = sys.platform.startswith("java")


def _create_default_type_map():
    if not _IS_JYTHON:
        default_type_map = [
            # None means that it should not be treated as a compound variable

            # isintance does not accept a tuple on some versions of python, so, we must declare it expanded
            (type(None), None,),
            (int, None),
            (float, None),
            (complex, None),
            (str, None),
            (tuple, pydevd_resolver.tupleResolver),
            (list, pydevd_resolver.tupleResolver),
            (dict, pydevd_resolver.dictResolver),
        ]
        try:
            default_type_map.append((long, None))  # @UndefinedVariable
        except:
            pass  # not available on all python versions

        try:
            default_type_map.append((unicode, None))  # @UndefinedVariable
        except:
            pass  # not available on all python versions

        try:
            default_type_map.append((set, pydevd_resolver.setResolver))
        except:
            pass  # not available on all python versions

        try:
            default_type_map.append((frozenset, pydevd_resolver.setResolver))
        except:
            pass  # not available on all python versions

        try:
            from django.utils.datastructures import MultiValueDict
            default_type_map.insert(0, (MultiValueDict, pydevd_resolver.multiValueDictResolver))
            # we should put it before dict
        except:
            pass  # django may not be installed

        try:
            from django.forms import BaseForm
            default_type_map.insert(0, (BaseForm, pydevd_resolver.djangoFormResolver))
            # we should put it before instance resolver
        except:
            pass  # django may not be installed

        try:
            from collections import deque
            default_type_map.append((deque, pydevd_resolver.dequeResolver))
        except:
            pass

        try:
            from collections import OrderedDict
            default_type_map.insert(0, (OrderedDict, pydevd_resolver.orderedDictResolver))
            # we should put it before dict
        except:
            pass

        if frame_type is not None:
            default_type_map.append((frame_type, pydevd_resolver.frameResolver))

    else:
        from org.python import core  # @UnresolvedImport
        default_type_map = [
            (core.PyNone, None),
            (core.PyInteger, None),
            (core.PyLong, None),
            (core.PyFloat, None),
            (core.PyComplex, None),
            (core.PyString, None),
            (core.PyTuple, pydevd_resolver.tupleResolver),
            (core.PyList, pydevd_resolver.tupleResolver),
            (core.PyDictionary, pydevd_resolver.dictResolver),
            (core.PyStringMap, pydevd_resolver.dictResolver),
        ]
        if hasattr(core, 'PyJavaInstance'):
            # Jython 2.5b3 removed it.
            default_type_map.append((core.PyJavaInstance, pydevd_resolver.instanceResolver))

    return default_type_map


class TypeResolveHandler(object):
    NO_PROVIDER = []  # Sentinel value (any mutable object to be used as a constant would be valid).

    def __init__(self):
        # Note: don't initialize with the types we already know about so that the extensions can override
        # the default resolvers that are already available if they want.
        self._type_to_resolver_cache = {}
        self._type_to_str_provider_cache = {}
        self._initialized = False

    def _initialize(self):
        self._default_type_map = _create_default_type_map()
        self._resolve_providers = pydevd_extension_utils.extensions_of_type(TypeResolveProvider)
        self._str_providers = pydevd_extension_utils.extensions_of_type(StrPresentationProvider)
        self._initialized = True

    def get_type(self, o):
        try:
            try:
                # Faster than type(o) as we don't need the function call.
                type_object = o.__class__
            except:
                # Not all objects have __class__ (i.e.: there are bad bindings around).
                type_object = type(o)

            type_name = type_object.__name__
        except:
            # This happens for org.python.core.InitModule
            return 'Unable to get Type', 'Unable to get Type', None

        return self._get_type(o, type_object, type_name)

    def _get_type(self, o, type_object, type_name):
        resolver = self._type_to_resolver_cache.get(type_object)
        if resolver is not None:
            return type_object, type_name, resolver

        if not self._initialized:
            self._initialize()

        try:
            for resolver in self._resolve_providers:
                if resolver.can_provide(type_object, type_name):
                    # Cache it
                    self._type_to_resolver_cache[type_object] = resolver
                    return type_object, type_name, resolver

            for t in self._default_type_map:
                if isinstance(o, t[0]):
                    # Cache it
                    resolver = t[1]
                    self._type_to_resolver_cache[type_object] = resolver
                    return (type_object, type_name, resolver)
        except:
            traceback.print_exc()

        # No match return default (and cache it).
        resolver = pydevd_resolver.defaultResolver
        self._type_to_resolver_cache[type_object] = resolver
        return type_object, type_name, resolver

    if _IS_JYTHON:
        _base_get_type = _get_type

        def _get_type(self, o, type_object, type_name):
            if type_name == 'org.python.core.PyJavaInstance':
                return type_object, type_name, pydevd_resolver.instanceResolver

            if type_name == 'org.python.core.PyArray':
                return type_object, type_name, pydevd_resolver.jyArrayResolver

            return self._base_get_type(o, type_name, type_name)

    def str_from_providers(self, o, type_object, type_name, do_trim=True):
        provider = self._type_to_str_provider_cache.get(type_object)

        if provider is self.NO_PROVIDER:
            return None

        if provider is not None:
            try:
                return provider.get_str(o, do_trim)
            except TypeError:
                return provider.get_str(o)

        if not self._initialized:
            self._initialize()

        for provider in self._str_providers:
            if provider.can_provide(type_object, type_name):
                self._type_to_str_provider_cache[type_object] = provider
                try:
                    return provider.get_str(o, do_trim)
                except TypeError:
                    return provider.get_str(o)

        self._type_to_str_provider_cache[type_object] = self.NO_PROVIDER
        return None


_TYPE_RESOLVE_HANDLER = TypeResolveHandler()

""" 
def get_type(o):
    Receives object and returns a triple (typeObject, typeString, resolver).

    resolver != None means that variable is a container, and should be displayed as a hierarchy.

    Use the resolver to get its attributes.

    All container objects should have a resolver.
"""
get_type = _TYPE_RESOLVE_HANDLER.get_type

_str_from_providers = _TYPE_RESOLVE_HANDLER.str_from_providers


def get_sorted_keys(frame_f_locals):
    keys = dict_keys(frame_f_locals)
    if hasattr(keys, 'sort'):
        keys.sort()  # Python 3.0 does not have it
    else:
        keys = sorted(keys)  # Jython 2.1 does not have it
    return keys


def frame_vars_to_xml(frame_f_locals, group_type, hidden_ns=None, user_type_renderers=None):
    """ dumps frame variables to XML
    <var name="var_name" scope="local" type="type" value="value"/>
    """
    keys = get_sorted_keys(frame_f_locals)

    type_handler = get_vars_handler(var_to_xml,
                                    handler_type=XML_COMMUNICATION_VARS_HANDLER,
                                    group_type=group_type)

    for k in keys:
        try:
            v = frame_f_locals[k]
            eval_full_val = should_evaluate_full_value(v, group_type)

            type_handler.handle(k, v, hidden_ns, eval_full_val, user_type_renderers=user_type_renderers)
        except Exception:
            traceback.print_exc()
            pydev_log.error("Unexpected error, recovered safely.\n")

    return type_handler.get_xml()


def _get_default_var_string_representation(v, _type, typeName, format, do_trim=True):
    str_from_provider = _str_from_providers(v, _type, typeName, do_trim)
    if str_from_provider is not None:
        return str_from_provider

    return get_value_repr(v, do_trim, format)


def var_to_xml(val, name, do_trim=True, additional_in_xml='', evaluate_full_value=True, format='%s', user_type_renderers=None):
    """ single variable or dictionary to xml representation """

    if name in DO_NOT_PROCESS_VARS:
        xml = '<var name="%s" ' % (make_valid_xml_value(name))
        return ''.join((xml, ' />\n'))

    try:
        # This should be faster than isinstance (but we have to protect against not having a '__class__' attribute).
        is_exception_on_eval = val.__class__ == ExceptionOnEvaluate
    except:
        is_exception_on_eval = False

    if is_exception_on_eval:
        v = val.result
    else:
        v = val

    _type, typeName, resolver = get_type(v)

    # type qualifier to xml
    type_qualifier = getattr(_type, "__module__", "")
    if type_qualifier:
        xml_qualifier = 'qualifier="%s"' % make_valid_xml_value(type_qualifier)
    else:
        xml_qualifier = ''

    # type renderer to xml
    type_renderer = None
    if user_type_renderers is not None:
        type_renderer = try_get_type_renderer_for_var(v, user_type_renderers)
    if type_renderer is not None:
        xml_type_renderer_id = 'typeRendererId="%s"' % make_valid_xml_value(type_renderer.to_type)
    else:
        xml_type_renderer_id = ''

    # name and type to xml
    try:
        name = _do_quote(name)  # TODO: Fix PY-5834 without using quote
    except:
        pass
    xml = '<var name="%s" type="%s" ' % (make_valid_xml_value(name), make_valid_xml_value(typeName))

    # value to xml
    value = None
    if not evaluate_full_value:
        value = DEFAULT_VALUES_DICT[LOAD_VALUES_POLICY]
    elif type_renderer is not None:
        value = type_renderer.evaluate_var_string_repr(v)
    if value is None:
        value = _get_default_var_string_representation(v, _type, typeName, format, do_trim)

    xml_value = ' value="%s"' % (make_valid_xml_value(_do_quote(value)))

    # shape to xml
    xml_shape = ''
    try:
        if should_evaluate_shape() and is_container_with_shape_dtype(type_qualifier, typeName, v):
            if hasattr(v, 'shape') and not callable(v.shape):
                xml_shape = ' shape="%s"' % make_valid_xml_value(str(tuple(v.shape)))
            elif hasattr(v, '__len__') and not is_string(v):
                xml_shape = ' shape="%s"' % make_valid_xml_value("%s" % str(len(v)))
    except:
        pass

    # data type info to xml (for arrays and tensors)
    # we use it for view as image
    xml_data_type = ''
    try:
        if (is_container_with_shape_dtype(type_qualifier, typeName, v)
                and hasattr(v.dtype, 'name')):
            xml_data_type = ' arrayElementType="%s"' % make_valid_xml_value(v.dtype.name)
    except:
        pass

    # additional info to xml
    if is_exception_on_eval:
        xml_container = ' isErrorOnEval="True"'
    else:
        if resolver is not None:
            xml_container = ' isContainer="True"'
        else:
            xml_container = ''

    return ''.join((xml, xml_qualifier, xml_value, xml_container, xml_shape, xml_data_type, xml_type_renderer_id, additional_in_xml, ' />\n'))


def _do_quote(elem):
    """ Quote the elem safely,
    e.g. encoding it first if necessary and decoding again after quoting.

    Note that strings in Python 2 are 'str' or 'unicode' types,
    'quote' function works only with 'str' and 'unicode' in ASCII types.
    In Python 3 all strings are unicode and have 'str' type,
    'quote' function works well with 'str' and 'bytes' types.

    :param elem: name or value of variable
    :type elem: str, unicode, bytes
    :return: elem in UTF-8
    :rtype: unicode, str
    """
    if not IS_PY3K:
        if elem.__class__.__name__ == 'unicode':
            # unicode to str
            result = elem.encode('utf-8')
            result = quote(result, '/>_= ')
            # str to unicode in UTF-8
            result = result.decode('utf-8')
        else:
            result = quote(elem, '/>_= ')
    else:
        result = quote(elem, '/>_= ')

    return result




try:
    import trace
except ImportError:
    pass
else:
    trace._warn = lambda *args: None   # workaround for http://bugs.python.org/issue17143 (PY-8706)

import os
from _pydevd_bundle.pydevd_comm import CMD_SIGNATURE_CALL_TRACE, NetCommand
from _pydevd_bundle import pydevd_xml
from _pydevd_bundle.pydevd_constants import xrange, dict_iter_items
from _pydevd_bundle import pydevd_utils
from _pydevd_bundle.pydevd_utils import get_clsname_for_code


class Signature(object):
    def __init__(self, file, name):
        self.file = file
        self.name = name
        self.args = []
        self.args_str = []
        self.return_type = None

    def add_arg(self, name, type):
        self.args.append((name, type))
        self.args_str.append("%s:%s"%(name, type))

    def set_args(self, frame, recursive=False):
        self.args = []

        code = frame.f_code
        locals = frame.f_locals

        for i in xrange(0, code.co_argcount):
            name = code.co_varnames[i]
            class_name = get_type_of_value(locals[name], recursive=recursive)

            self.add_arg(name, class_name)

    def __str__(self):
        return "%s %s(%s)"%(self.file, self.name, ", ".join(self.args_str))


def get_type_of_value(value, ignore_module_name=('__main__', '__builtin__', 'builtins'), recursive=False):
    tp = type(value)
    class_name = tp.__name__
    if class_name == 'instance':  # old-style classes
        tp = value.__class__
        class_name = tp.__name__

    if hasattr(tp, '__module__') and tp.__module__ and tp.__module__ not in ignore_module_name:
        class_name = "%s.%s"%(tp.__module__, class_name)

    if class_name == 'list':
        class_name = 'List'
        if len(value) > 0 and recursive:
            class_name += '[%s]' % get_type_of_value(value[0], recursive=recursive)
        return class_name

    if class_name == 'dict':
        class_name = 'Dict'
        if len(value) > 0 and recursive:
            for (k, v) in dict_iter_items(value):
                class_name += '[%s, %s]' % (get_type_of_value(k, recursive=recursive), 
                                            get_type_of_value(v, recursive=recursive))
                break
        return class_name

    if class_name == 'tuple':
        class_name = 'Tuple'
        if len(value) > 0 and recursive:
            class_name += '['
            class_name += ', '.join(get_type_of_value(v, recursive=recursive) for v in value)
            class_name += ']'

    return class_name


def _modname(path):
    """Return a plausible module name for the path"""
    base = os.path.basename(path)
    filename, ext = os.path.splitext(base)
    return filename


class SignatureFactory(object):
    def __init__(self):
        self._caller_cache = {}
        self.cache = CallSignatureCache()

    def is_in_scope(self, filename):
        return pydevd_utils.in_project_roots(filename)

    def create_signature(self, frame, filename, with_args=True):
        try:
            _, modulename, funcname = self.file_module_function_of(frame)
            signature = Signature(filename, funcname)
            if with_args:
                signature.set_args(frame, recursive=True)
            return signature
        except:
            import traceback
            traceback.print_exc()


    def file_module_function_of(self, frame): #this code is take from trace module and fixed to work with new-style classes
        code = frame.f_code
        filename = code.co_filename
        if filename:
            modulename = _modname(filename)
        else:
            modulename = None

        funcname = code.co_name
        clsname = None
        if code in self._caller_cache:
            if self._caller_cache[code] is not None:
                clsname = self._caller_cache[code]
        else:
            self._caller_cache[code] = None
            clsname = get_clsname_for_code(code, frame)
            if clsname is not None:
                # cache the result - assumption is that new.* is
                # not called later to disturb this relationship
                # _caller_cache could be flushed if functions in
                # the new module get called.
                self._caller_cache[code] = clsname

        if clsname is not None:
            funcname = "%s.%s" % (clsname, funcname)

        return filename, modulename, funcname


def get_signature_info(signature):
    return signature.file, signature.name, ' '.join([arg[1] for arg in signature.args])


def get_frame_info(frame):
    co = frame.f_code
    return co.co_name, frame.f_lineno, co.co_filename


class CallSignatureCache(object):
    def __init__(self):
        self.cache = {}

    def add(self, signature):
        filename, name, args_type = get_signature_info(signature)
        calls_from_file = self.cache.setdefault(filename, {})
        name_calls = calls_from_file.setdefault(name, {})
        name_calls[args_type] = None

    def is_in_cache(self, signature):
        filename, name, args_type = get_signature_info(signature)
        if args_type in self.cache.get(filename, {}).get(name, {}):
            return True
        return False


def create_signature_message(signature):
    cmdTextList = ["<xml>"]

    cmdTextList.append('<call_signature file="%s" name="%s">' % (pydevd_xml.make_valid_xml_value(signature.file), pydevd_xml.make_valid_xml_value(signature.name)))

    for arg in signature.args:
        cmdTextList.append('<arg name="%s" type="%s"></arg>' % (pydevd_xml.make_valid_xml_value(arg[0]), pydevd_xml.make_valid_xml_value(arg[1])))
        
    if signature.return_type is not None:
        cmdTextList.append('<return type="%s"></return>' % (pydevd_xml.make_valid_xml_value(signature.return_type)))

    cmdTextList.append("</call_signature></xml>")
    cmdText = ''.join(cmdTextList)
    return NetCommand(CMD_SIGNATURE_CALL_TRACE, 0, cmdText)


def send_signature_call_trace(dbg, frame, filename):
    if dbg.signature_factory and dbg.signature_factory.is_in_scope(filename):
        signature = dbg.signature_factory.create_signature(frame, filename)
        if signature is not None:
            if dbg.signature_factory.cache is not None:
                if not dbg.signature_factory.cache.is_in_cache(signature):
                    dbg.signature_factory.cache.add(signature)
                    dbg.writer.add_command(create_signature_message(signature))
                    return True
                else:
                    # we don't send signature if it is cached
                    return False
            else:
                dbg.writer.add_command(create_signature_message(signature))
                return True
    return False


def send_signature_return_trace(dbg, frame, filename, return_value):
    if dbg.signature_factory and dbg.signature_factory.is_in_scope(filename):
        signature = dbg.signature_factory.create_signature(frame, filename, with_args=False)
        signature.return_type = get_type_of_value(return_value, recursive=True)
        dbg.writer.add_command(create_signature_message(signature))
        return True

    return False






from _pydevd_bundle.pydevd_constants import dict_keys, dict_iter_items

try:
    from code import InteractiveConsole, InteractiveInterpreter
except ImportError:
    from _pydevd_bundle.pydevconsole_code_for_ironpython import IronPythonInteractiveConsole as InteractiveConsole, \
        IronPythonInteractiveInterpreter as InteractiveInterpreter

import os
import sys
import traceback
from code import compile_command

from _pydev_bundle.pydev_code_executor import BaseCodeExecutor
from _pydev_bundle.pydev_console_types import CodeFragment, Command
from _pydev_bundle.pydev_imports import Exec
from _pydevd_bundle import pydevd_vars, pydevd_save_locals
from _pydevd_bundle.pydevd_console_pytest import enable_pytest_output
from _pydevd_bundle.pydevd_asyncio_provider import get_asyncio_command_compiler, get_exec_async_code, get_apply

try:
    import __builtin__
except:
    import builtins as __builtin__  # @UnresolvedImport


class CodeExecutor(BaseCodeExecutor):
    def __init__(self):
        super(CodeExecutor, self).__init__()

        self.namespace = {}
        self.interpreter = InteractiveConsole(self.namespace)

    def do_add_exec(self, codeFragment):
        command = Command(self.interpreter, codeFragment)
        command.run()
        return command.more

    def get_namespace(self):
        return self.namespace


if 'IPYTHONENABLE' in os.environ:
    IPYTHON = os.environ['IPYTHONENABLE'] == 'True'
else:
    IPYTHON = True

PYTEST_RUN_CONFIG = 'PYTEST_RUN_CONFIG' in os.environ

try:
    try:
        exitfunc = sys.exitfunc
    except AttributeError:
        exitfunc = None

    if IPYTHON:
        from _pydev_bundle.pydev_ipython_code_executor import IPythonCodeExecutor as CodeExecutor
        if exitfunc is not None:
            sys.exitfunc = exitfunc
        else:
            try:
                delattr(sys, 'exitfunc')
            except:
                pass
except:
    IPYTHON = False
    pass


def get_ipython_hidden_vars():
    if IPYTHON and hasattr(__builtin__, 'interpreter'):
        code_executor = get_code_executor()
        return code_executor.get_ipython_hidden_vars_dict()
    else:
        try:
            ipython_shell = get_ipython()
            from _pydev_bundle.pydev_ipython_console_011 import get_ipython_hidden_vars
            return get_ipython_hidden_vars(ipython_shell)
        except:
            pass


def get_code_executor():
    try:
        code_executor = getattr(__builtin__, 'interpreter')
    except AttributeError:
        code_executor = CodeExecutor()
        __builtin__.interpreter = code_executor
        print(code_executor.get_greeting_msg())

    return code_executor


# ===============================================================================
# Debugger integration
# ===============================================================================

def ipython_exec_code(code, globals, locals, debugger):
    code_executor = get_code_executor()
    code_executor.interpreter.update(globals, locals)

    res = code_executor.need_more(code)

    if res:
        return True, False

    more, exception_occurred = code_executor.add_exec(code, debugger)

    ipython = code_executor.interpreter.ipython
    for key in dict_keys(ipython.user_ns):
        locals[key] = ipython.user_ns[key]
    for key in dict_keys(locals):
        if key not in ipython.user_ns:
            locals.pop(key)

    return more, exception_occurred


class ConsoleWriter(InteractiveInterpreter):
    skip = 0

    def __init__(self, locals=None):
        InteractiveInterpreter.__init__(self, locals)

    def write(self, data):
        # if (data.find("global_vars") == -1 and data.find("pydevd") == -1):
        if self.skip > 0:
            self.skip -= 1
        else:
            if data == "Traceback (most recent call last):\n":
                self.skip = 1
            sys.stderr.write(data)

    def showsyntaxerror(self, filename=None):
        """Display the syntax error that just occurred."""
        # Override for avoid using sys.excepthook PY-12600
        type, value, tb = sys.exc_info()
        sys.last_type = type
        sys.last_value = value
        sys.last_traceback = tb
        if filename and type is SyntaxError:
            # Work hard to stuff the correct filename in the exception
            try:
                msg, (dummy_filename, lineno, offset, line) = value.args
            except ValueError:
                # Not the format we expect; leave it alone
                pass
            else:
                # Stuff in the right filename
                value = SyntaxError(msg, (filename, lineno, offset, line))
                sys.last_value = value
        list = traceback.format_exception_only(type, value)
        sys.stderr.write(''.join(list))

    def showtraceback(self):
        """Display the exception that just occurred."""
        # Override for avoid using sys.excepthook PY-12600
        try:
            type, value, tb = sys.exc_info()
            sys.last_type = type
            sys.last_value = value
            sys.last_traceback = tb
            tblist = traceback.extract_tb(tb)
            del tblist[:1]
            lines = traceback.format_list(tblist)
            if lines:
                lines.insert(0, "Traceback (most recent call last):\n")
            lines.extend(traceback.format_exception_only(type, value))
        finally:
            tblist = tb = None
        sys.stderr.write(''.join(lines))


def console_exec(thread_id, frame_id, expression, dbg):
    """returns 'False' in case expression is partially correct
    """
    frame = pydevd_vars.find_frame(thread_id, frame_id)

    try:
        expression = str(expression.replace('@LINE@', '\n'))
    except UnicodeEncodeError as e:
        expression = expression.replace('@LINE@', '\n')

    # Not using frame.f_globals because of https://sourceforge.net/tracker2/?func=detail&aid=2541355&group_id=85796&atid=577329
    # (Names not resolved in generator expression in method)
    # See message: http://mail.python.org/pipermail/python-list/2009-January/526522.html
    updated_globals = {}
    updated_globals.update(frame.f_globals)
    updated_globals.update(frame.f_locals)  # locals later because it has precedence over the actual globals

    if PYTEST_RUN_CONFIG:
        enable_pytest_output()

    if IPYTHON:
        apply_func = get_apply()
        if apply_func is not None:
            apply_func()
        need_more, exception_occurred = ipython_exec_code(CodeFragment(expression), updated_globals, updated_globals, dbg)
        if not need_more:
            update_frame_local_variables_and_save(frame, updated_globals)
        return need_more, exception_occurred

    interpreter = ConsoleWriter()

    try:
        async_compiler = get_asyncio_command_compiler()
        if async_compiler is not None:
            compiler = async_compiler
        else:
            compiler = compile_command
        try:
            code = compiler(expression)
        except (OverflowError, SyntaxError, ValueError):
            code = compiler(expression, symbol='exec')
    except (OverflowError, SyntaxError, ValueError):
        # Case 1
        interpreter.showsyntaxerror()
        return False, True
    if code is None:
        # Case 2
        return True, False

    # Case 3
    code_executor = get_code_executor()
    code_executor.interruptable = True
    exception_occurred = False
    try:
        exec_func = get_exec_async_code()
        if exec_func is not None:
            exec_func(code, updated_globals)
        else:
            # It is important that globals and locals we pass to the exec function are the same object.
            # Otherwise generator expressions can confuse their scope. Passing updated_globals dictionary seems to be a safe option here
            # because it contains globals and locals in the right precedence.
            # See: https://stackoverflow.com/questions/15866398/why-is-a-python-generator-confusing-its-scope-with-global-in-an-execd-script.
            Exec(code, updated_globals, updated_globals)
    except:
        interpreter.showtraceback()
        exception_occurred = True
    else:
        update_frame_local_variables_and_save(frame, updated_globals)
    finally:
        code_executor.interruptable = False
    return False, exception_occurred


def interrupt_debug_console():
    code_executor = get_code_executor()
    code_executor.interrupt()


def update_frame_local_variables_and_save(frame, values):
    """Update the frame local variables with the values from `values`, remove those that no longer exist, and save."""
    # It is important to access to `frame.f_locals` through a local reference.
    # Any changes done to it will be lost as soon as we access the attribute next time.
    f_locals, f_globals = frame.f_locals, frame.f_globals

    for key, value in dict_iter_items(values):
        if key not in f_globals or value is not f_globals[key]:
            f_locals[key] = value

    for key in dict_keys(f_locals):
        if key not in values:
            f_locals.pop(key)

    pydevd_save_locals.save_locals(frame)



#  Copyright 2000-2025 JetBrains s.r.o. and contributors. Use of this source code is governed by the Apache 2.0 license.
import base64
import io
import uuid

IMAGE_DATA_STORAGE = {}
DEFAULT_IMAGE_FORMAT = 'PNG'
DEFAULT_ENCODING = 'utf-8'
GRAYSCALE_MODE = 'L'
RGB_MODE = 'RGB'
RGBA_MODE = 'RGBA'
CHUNK_SIZE = 8192

def load_image_chunk(offset, image_id):
    # type: (int, str) -> str
    try:
        bytes_data = IMAGE_DATA_STORAGE.get(image_id)
        if bytes_data is None:
            return "Error: No image data found."
        chunk = bytes_data[offset:offset + CHUNK_SIZE]
        next_offset = offset + CHUNK_SIZE
        if next_offset >= len(bytes_data):
            next_offset = -1
            IMAGE_DATA_STORAGE.pop(image_id, None)
        chunk_bytes = base64.b64encode(chunk)
        if not isinstance(chunk_bytes, str):
            chunk_bytes = chunk_bytes.decode(DEFAULT_ENCODING)
        return "{};{}".format(chunk_bytes, next_offset)
    except ValueError:
        return "Error: Invalid offset format."
    except Exception as e:
        return "Error: {}".format(e)


def save_image_to_storage(image_data, data_type=None, format=DEFAULT_IMAGE_FORMAT, save_func=None):
    # type: (any, str, str, callable) -> str
    try:
        bytes_buffer = io.BytesIO()
        try:
            if save_func:
                save_func(bytes_buffer, format)
            else:
                image_data.save(bytes_buffer, format=format)
            bytes_buffer.seek(0)
            bytes_data = bytes_buffer.getvalue()
            image_id = str(uuid.uuid4())
            IMAGE_DATA_STORAGE[image_id] = bytes_data
            if data_type is None:
                data_type = "None"
            return "{};{};{}".format(image_id, len(bytes_data), data_type)
        finally:
            bytes_buffer.close()
    except Exception as e:
        return "Error: {}".format(e)



from __future__ import print_function

# Important: Autogenerated file.

# DO NOT edit manually!
# DO NOT edit manually!
# Copyright: Brainwy Software
#
# License: EPL

import os
import sys
import threading
import traceback
from os.path import splitext, basename

from _pydev_bundle import pydev_log
from _pydevd_bundle.pydevd_breakpoints import stop_on_unhandled_exception
from _pydevd_bundle.pydevd_bytecode_utils import (
    find_last_call_name, find_last_func_call_order)
from _pydevd_bundle.pydevd_comm_constants import (
    CMD_STEP_OVER, CMD_SMART_STEP_INTO, CMD_SET_BREAK, CMD_STEP_INTO,
    CMD_STEP_INTO_MY_CODE, CMD_STEP_INTO_COROUTINE, CMD_STEP_RETURN)
from _pydevd_bundle.pydevd_constants import (
    PYDEVD_TOOL_NAME, STATE_RUN, STATE_SUSPEND, GlobalDebuggerHolder, IS_CPYTHON, IS_PY313_OR_GREATER)
from _pydevd_bundle.pydevd_dont_trace_files import DONT_TRACE, PYDEV_FILE
from _pydevd_bundle.pydevd_trace_dispatch import (
    set_additional_thread_info, handle_breakpoint_condition,
    handle_breakpoint_expression, DEBUG_START, DEBUG_START_PY3K,
    should_stop_on_exception, handle_exception, manage_return_values)
from pydevd_file_utils import (
    NORM_PATHS_AND_BASE_CONTAINER, get_abs_path_real_path_and_base_from_frame)

get_file_type = DONT_TRACE.get

global_cache_skips = {}
global_cache_frame_skips = {}

try:
    monitoring = sys.monitoring
except AttributeError:
    pass

_EVENT_ACTIONS = {
    "ADD": lambda x, y: x | y,
    "REMOVE": lambda x, y: x & ~y,
}

try:
    _thread_local_info = threading.local()
    _get_ident = threading.get_ident
    _thread_active = threading._active  # noqa
except:
    pass


def _get_bootstrap_frame(depth):
    try:
        return _thread_local_info.f_bootstrap, _thread_local_info.is_bootstrap_frame_internal
    except:
        frame = _getframe(depth)
        f_bootstrap = frame
        # print('called at', f_bootstrap.f_code.co_name, f_bootstrap.f_code.co_filename, f_bootstrap.f_code.co_firstlineno)
        is_bootstrap_frame_internal = False
        while f_bootstrap is not None:
            filename = f_bootstrap.f_code.co_filename
            name = splitext(basename(filename))[0]

            if name == "threading":
                if f_bootstrap.f_code.co_name in ("__bootstrap", "_bootstrap"):
                    # We need __bootstrap_inner, not __bootstrap.
                    return None, False

                elif f_bootstrap.f_code.co_name in ("__bootstrap_inner", "_bootstrap_inner", "is_alive"):
                    # Note: be careful not to use threading.current_thread to avoid creating a dummy thread.
                    is_bootstrap_frame_internal = True
                    break

            elif name == "pydev_monkey":
                if f_bootstrap.f_code.co_name == "__call__":
                    is_bootstrap_frame_internal = True
                    break

            elif name == "pydevd":
                if f_bootstrap.f_code.co_name in ("run", "main"):
                    # We need to get to _exec
                    return None, False

                if f_bootstrap.f_code.co_name == "_exec":
                    is_bootstrap_frame_internal = True
                    break

            elif f_bootstrap.f_back is None:
                break

            f_bootstrap = f_bootstrap.f_back

        if f_bootstrap is not None:
            _thread_local_info.is_bootstrap_frame_internal = is_bootstrap_frame_internal
            _thread_local_info.f_bootstrap = f_bootstrap
            return _thread_local_info.f_bootstrap, _thread_local_info.is_bootstrap_frame_internal

        return f_bootstrap, is_bootstrap_frame_internal


class ThreadInfo:
    def __init__(self, thread, thread_ident, trace, additional_info):
        self.thread = thread
        self.thread_ident = thread_ident
        self.additional_info = additional_info
        self.trace = trace

        self._use_handle = hasattr(thread, "_handle")
        self._use_started = hasattr(thread, "_started")
        self._use_os_thread_handle = hasattr(thread, "_os_thread_handle")

    def is_thread_alive(self):
        # Python >=3.14
        if self._use_os_thread_handle and self._use_started:
            return not self.thread._os_thread_handle.is_done()

        # Python ==3.13
        elif self._use_handle and self._use_started:
            return not self.thread._handle.is_done()

        # Python ==3.12
        else:
            return not self.thread._is_stopped


class _DeleteDummyThreadOnDel:
    """
    Helper class to remove a dummy thread from threading._active on __del__.
    """

    def __init__(self, dummy_thread):
        self._dummy_thread = dummy_thread
        self._tident = dummy_thread.ident
        # Put the thread on a thread local variable so that when
        # the related thread finishes this instance is collected.
        #
        # Note: no other references to this instance may be created.
        # If any client code creates a reference to this instance,
        # the related _DummyThread will be kept forever!
        _thread_local_info._track_dummy_thread_ref = self

    def __del__(self):
        with threading._active_limbo_lock:
            if _thread_active.get(self._tident) is self._dummy_thread:
                _thread_active.pop(self._tident, None)


def _create_thread_info(depth):
    # Don't call threading.currentThread because if we're too early in the process
    # we may create a dummy thread.
    thread_ident = _get_ident()

    f_bootstrap_frame, is_bootstrap_frame_internal = _get_bootstrap_frame(depth + 1)
    if f_bootstrap_frame is None:
        return None  # Case for threading when it's still in bootstrap or early in pydevd.

    if is_bootstrap_frame_internal:
        t = None
        if f_bootstrap_frame.f_code.co_name in ("__bootstrap_inner", "_bootstrap_inner", "is_alive"):
            # Note: be careful not to use threading.current_thread to avoid creating a dummy thread.
            t = f_bootstrap_frame.f_locals.get("self")
            if not isinstance(t, threading.Thread):
                t = None

        elif f_bootstrap_frame.f_code.co_name in ("_exec", "__call__"):
            # Note: be careful not to use threading.current_thread to avoid creating a dummy thread.
            t = f_bootstrap_frame.f_locals.get("t")
            if not isinstance(t, threading.Thread):
                t = None

    else:
        # This means that the first frame is not in threading nor in pydevd.
        # In practice this means it's some unmanaged thread, so, creating
        # a dummy thread is ok in this use-case.
        t = threading.current_thread()

    if t is None:
        t = _thread_active.get(thread_ident)

    if isinstance(t, threading._DummyThread) and not IS_PY313_OR_GREATER:
        _thread_local_info._ref = _DeleteDummyThreadOnDel(t)

    if t is None:
        return None

    if getattr(t, "is_pydev_daemon_thread", False):
        return ThreadInfo(t, thread_ident, False, None)
    else:
        try:
            additional_info = t.additional_info
            if additional_info is None:
                raise AttributeError()
        except:
            additional_info = set_additional_thread_info(t)
        return ThreadInfo(t, thread_ident, True, additional_info)

def _get_thread_info(create, depth):
    """
    Provides thread-related info.

    May return None if the thread is still not active.
    """
    try:
        # Note: changing to a `dict[thread.ident] = thread_info` had almost no
        # effect in the performance.
        return _thread_local_info.thread_info
    except:
        if not create:
            return None
        thread_info = _create_thread_info(depth + 1)
        if thread_info is None:
            return None

        _thread_local_info.thread_info = thread_info
        return _thread_local_info.thread_info

def _make_frame_cache_key(code):
    return code.co_firstlineno, code.co_name, code.co_filename


def _get_additional_info(thread):
    try:
        additional_info = thread.additional_info
        if additional_info is None:
            raise AttributeError()
    except:
        additional_info = set_additional_thread_info(thread)
    return additional_info


def _get_abs_path_real_path_and_base_from_frame(frame):
    try:
        abs_path_real_path_and_base = NORM_PATHS_AND_BASE_CONTAINER[
            frame.f_code.co_filename]
    except:
        abs_path_real_path_and_base \
            = get_abs_path_real_path_and_base_from_frame(frame)

    return abs_path_real_path_and_base


def _should_enable_line_events_for_code(frame, code, filename, info, will_be_stopped=False):
    line_number = frame.f_lineno

    # print('PY_START (should enable line events check) %s %s %s %s' % (line_number, code.co_name, filename, info.pydev_step_cmd))

    py_db = GlobalDebuggerHolder.global_dbg
    if py_db is None:
        return monitoring.DISABLE

    plugin_manager = py_db.plugin

    if info is None:
        return False

    stop_frame = info.pydev_step_stop
    step_cmd = info.pydev_step_cmd

    breakpoints_for_file = py_db.breakpoints.get(filename)

    can_skip = False

    if info.pydev_state == 1 and not will_be_stopped:  # STATE_RUN = 1
        can_skip = (step_cmd == -1 and stop_frame is None) \
                   or (step_cmd in (109, 108) and stop_frame is not frame)

        if can_skip:
            if plugin_manager is not None and py_db.has_plugin_line_breaks:
                can_skip = not plugin_manager.can_not_skip(py_db, frame, info)

            # CMD_STEP_OVER = 108
            if (can_skip and py_db.show_return_values
                    and info.pydev_step_cmd == 108
                    and frame.f_back is info.pydev_step_stop):
                # trace function for showing return values after step over
                can_skip = False

    frame_cache_key = _make_frame_cache_key(code)
    line_cache_key = (frame_cache_key, line_number)

    if breakpoints_for_file:
        # When cached, 0 means we don't have a breakpoint
        # and 1 means we have.
        if can_skip:
            breakpoints_in_line_cache = global_cache_frame_skips.get(line_cache_key, -1)
            if breakpoints_in_line_cache == 0:
                return False

        breakpoints_in_frame_cache = global_cache_frame_skips.get(frame_cache_key, -1)
        if breakpoints_in_frame_cache != -1:
            # Gotten from cache.
            has_breakpoint_in_frame = breakpoints_in_frame_cache == 1
        else:
            has_breakpoint_in_frame = False
            # Checks the breakpoint to see if there is a context
            # match in some function.
            curr_func_name = frame.f_code.co_name

            # global context is set with an empty name
            if curr_func_name in ('?', '<module>', '<lambda>'):
                curr_func_name = ''

            for breakpoint in breakpoints_for_file.values():
                # will match either global or some function
                if breakpoint.func_name in ('None', curr_func_name):
                    has_breakpoint_in_frame = True
                    # New breakpoint was processed -> stop tracing monitoring.events.INSTRUCTION
                    remove_breakpoint(breakpoint)
                    break

                # Check is f_back has a breakpoint => need register return event
                if hasattr(frame, "f_back"):
                    f_code = getattr(frame.f_back, "f_code", None)
                    if f_code is not None and breakpoint.func_name == f_code.co_name:
                        can_skip = False
                        break

            # Cache the value (1 or 0 or -1 for default because of cython).
            if has_breakpoint_in_frame:
                global_cache_frame_skips[frame_cache_key] = 1
            else:
                global_cache_frame_skips[frame_cache_key] = 0

        if can_skip and not has_breakpoint_in_frame:
            return False

    return True


def _clear_run_state(info):
    if info is None:
        return

    info.pydev_step_stop = None
    info.pydev_step_cmd = -1
    info.pydev_state = STATE_RUN

# Cythonnized functions live in C the call stack rather than in the Python call stack.
# It means that the function gets its Python caller's stack frame at depth 0, and
# not its own.
# This behavior should be treated as undefined, and can be changed on the
# Cython side in the future.
# IFDEF CYTHON -- DONT EDIT THIS FILE (it is automatically generated)
cdef _getframe(depth=0):
   return sys._getframe()
# ELSE
# _getframe = sys._getframe
# ENDIF


def _get_top_level_frame():
    f_unhandled = _getframe()

    while f_unhandled:
        filename = f_unhandled.f_code.co_filename
        name = splitext(basename(filename))[0]
        if name == 'pydevd':
            if f_unhandled.f_code.co_name == '_exec':
                break
        elif name == 'threading':
            if f_unhandled.f_code.co_name == '_bootstrap_inner':
                break
        f_unhandled = f_unhandled.f_back

    return f_unhandled


def _stop_on_unhandled_exception(exc_info, py_db, thread):
    additional_info = _get_additional_info(thread)
    if additional_info is None:
        return

    if not additional_info.suspended_at_unhandled:
        additional_info.suspended_at_unhandled = True
        stop_on_unhandled_exception(py_db, thread, additional_info,
                                    exc_info)


def enable_pep669_monitoring():
    DEBUGGER_ID = monitoring.DEBUGGER_ID
    if not monitoring.get_tool(DEBUGGER_ID):
        monitoring.use_tool_id(DEBUGGER_ID, PYDEVD_TOOL_NAME)

        monitoring.set_events(
            DEBUGGER_ID,
            monitoring.events.PY_START | monitoring.events.RAISE
        )

        for event_type, callback in (
            (monitoring.events.PY_START, py_start_callback),
            (monitoring.events.LINE, py_line_callback),
            (monitoring.events.PY_RETURN, py_return_callback),
            (monitoring.events.RAISE, py_raise_callback),
            (monitoring.events.CALL, call_callback),
        ):
            monitoring.register_callback(DEBUGGER_ID, event_type, callback)

    debugger = GlobalDebuggerHolder.global_dbg
    if debugger:
        debugger.is_pep669_monitoring_enabled = True


def add_new_breakpoint(breakpoint):
    breakpoint._not_processed = True
    monitoring.restart_events()
    _modify_global_events(_EVENT_ACTIONS["ADD"], monitoring.events.CALL)


def remove_breakpoint(breakpoint):
    if getattr(breakpoint, '_not_processed', None):
        breakpoint._not_processed = False
        _modify_global_events(_EVENT_ACTIONS["REMOVE"], monitoring.events.CALL)


def _modify_global_events(action, event):
    DEBUGGER_ID = monitoring.DEBUGGER_ID
    if not monitoring.get_tool(DEBUGGER_ID):
        return

    current_events = monitoring.get_events(DEBUGGER_ID)
    monitoring.set_events(DEBUGGER_ID, action(current_events, event))


def _enable_return_tracing(code):
    local_events = monitoring.get_local_events(monitoring.DEBUGGER_ID, code)
    monitoring.set_local_events(monitoring.DEBUGGER_ID, code,
                                local_events | monitoring.events.PY_RETURN)


def _enable_line_tracing(code):
    local_events = monitoring.get_local_events(monitoring.DEBUGGER_ID, code)
    monitoring.set_local_events(monitoring.DEBUGGER_ID, code,
                                local_events | monitoring.events.LINE)


def call_callback(code, instruction_offset, callable, arg0):
    try:
        py_db = GlobalDebuggerHolder.global_dbg
    except AttributeError:
        return
    if py_db is None:
        return monitoring.DISABLE

    frame = _getframe(1)
    # print('ENTER: CALL ', code.co_filename, frame.f_lineno, code.co_name)

    try:
        if py_db.pydb_disposed:
            return monitoring.DISABLE

        try:
            thread_info = _thread_local_info.thread_info
        except:
            thread_info = _get_thread_info(True, 1)
            if thread_info is None:
                return


        if not thread_info.is_thread_alive():
            return

        frame_cache_key = _make_frame_cache_key(code)

        info = thread_info.additional_info
        if info is None:
            return

        pydev_step_cmd = info.pydev_step_cmd
        is_stepping = pydev_step_cmd != -1

        if not is_stepping and frame_cache_key in global_cache_skips:
            return monitoring.DISABLE

        abs_path_real_path_and_base = _get_abs_path_real_path_and_base_from_frame(frame)
        filename = abs_path_real_path_and_base[1]

        breakpoints_for_file = (py_db.breakpoints.get(filename)
                                or py_db.has_plugin_line_breaks)
        if not breakpoints_for_file and not is_stepping:
            return monitoring.DISABLE

        if _should_enable_line_events_for_code(frame, code, filename, info):
            _enable_line_tracing(code)
            _enable_return_tracing(code)
    except SystemExit:
        return monitoring.DISABLE
    except Exception:
        try:
            if traceback is not None:
                traceback.print_exc()
        except:
            pass
        return monitoring.DISABLE


def py_start_callback(code, instruction_offset):
    try:
        py_db = GlobalDebuggerHolder.global_dbg
    except AttributeError:
        return

    if py_db is None:
        return monitoring.DISABLE

    frame = _getframe(1)

    # print('ENTER: PY_START ', code.co_filename, frame.f_lineno, code.co_name)

    try:
        thread_info = _thread_local_info.thread_info
    except:
        thread_info = _get_thread_info(True, 1)
        if thread_info is None:
            return

    try:
        if py_db.pydb_disposed:
            return monitoring.DISABLE

        if not thread_info.trace or not thread_info.is_thread_alive():
            # For thread-related stuff we can't disable the code tracing because other
            # threads may still want it...
            return

        if py_db.thread_analyser is not None:
            py_db.thread_analyser.log_event(frame)

        if py_db.asyncio_analyser is not None:
            py_db.asyncio_analyser.log_event(frame)

        frame_cache_key = _make_frame_cache_key(code)

        info = thread_info.additional_info
        if info is None:
            return

        pydev_step_cmd = info.pydev_step_cmd
        is_stepping = pydev_step_cmd != -1

        if not is_stepping and frame_cache_key in global_cache_skips:
            # print('skipped: PY_START (cache hit)', frame_cache_key, frame.f_lineno, code.co_name)
            return

        abs_path_real_path_and_base = _get_abs_path_real_path_and_base_from_frame(frame)
        filename = abs_path_real_path_and_base[1]
        file_type = get_file_type(abs_path_real_path_and_base[-1])

        if file_type is not None:
            if file_type == 1:  # inlining LIB_FILE = 1
                if not py_db.in_project_scope(filename):
                    # print('skipped: PY_START (not in scope)', abs_path_real_path_and_base[-1], frame.f_lineno, code.co_name, file_type)
                    global_cache_skips[frame_cache_key] = 1
                    return monitoring.DISABLE
            else:
                # print('skipped: PY_START', abs_path_real_path_and_base[-1], frame.f_lineno, code.co_name, file_type)
                global_cache_skips[frame_cache_key] = 1
                return monitoring.DISABLE

        breakpoints_for_file = (py_db.breakpoints.get(filename)
                                or py_db.has_plugin_line_breaks)
        if not breakpoints_for_file and not is_stepping:
            return

        if py_db.plugin and py_db.has_plugin_line_breaks:
            args = (py_db, filename, info, thread_info.thread)
            result = py_db.plugin.get_breakpoint(py_db, frame, 'call', args)
            if result is not None:
                flag, breakpoint, new_frame, bp_type = result
                if breakpoint:
                    eval_result = False
                    if breakpoint.has_condition:
                        eval_result = handle_breakpoint_condition(py_db, info, breakpoint, new_frame)

                    if breakpoint.expression is not None:
                        handle_breakpoint_expression(breakpoint, info, new_frame)
                        if breakpoint.is_logpoint and info.pydev_message is not None and len(
                                info.pydev_message) > 0:
                            cmd = py_db.cmd_factory.make_io_message(info.pydev_message + os.linesep, '1')
                            py_db.writer.add_command(cmd)

                    if breakpoint.has_condition and not eval_result:
                        return

                if flag:
                    result = py_db.plugin.suspend(py_db, thread_info.thread, frame, bp_type)
                    if result is not None:
                        frame = result

                if info.pydev_state == STATE_SUSPEND:
                    py_db.do_wait_suspend(thread_info.thread, frame, 'call', None)
                    return

        if is_stepping:
            if (pydev_step_cmd == CMD_STEP_OVER
                    and frame is not info.pydev_step_stop):
                if frame.f_back is info.pydev_step_stop:
                    _enable_return_tracing(code)
            if (py_db.is_filter_enabled
                    and py_db.is_ignored_by_filters(filename)):
                return monitoring.DISABLE
            if (py_db.is_filter_libraries
                    and not py_db.in_project_scope(filename)):
                return monitoring.DISABLE
            # We are stepping, and there is no reason to skip the frame
            # at this point.
            _enable_line_tracing(code)
            _enable_return_tracing(code)

            # Process plugins stepping
            stop_info = {}
            args = (py_db, filename, info, thread_info.thread)
            stop_frame = info.pydev_step_stop
            plugin_stop = False

            if py_db.plugin is not None:
                if pydev_step_cmd == CMD_STEP_INTO:
                    result = py_db.plugin.cmd_step_into(py_db, frame, 'call', args, stop_info, False)
                    if result:
                        _, plugin_stop = result
                if pydev_step_cmd == CMD_STEP_OVER:
                    result = py_db.plugin.cmd_step_over(py_db, frame, 'call', args, stop_info, stop_frame is frame)
                    if result:
                        _, plugin_stop = result

            if plugin_stop:
                py_db.plugin.stop(py_db, frame, 'call', args, stop_info, None, pydev_step_cmd)

        if _should_enable_line_events_for_code(frame, code, filename, info):
            _enable_line_tracing(code)
            _enable_return_tracing(code)
        else:
            global_cache_skips[frame_cache_key] = 1
            return

    except SystemExit:
        return monitoring.DISABLE
    except Exception:
        try:
            if traceback is not None:
                traceback.print_exc()
        except:
            pass
        return monitoring.DISABLE


def py_line_callback(code, line_number):
    frame = _getframe(1)

    try:
        thread_info = _thread_local_info.thread_info
    except:
        thread_info = _get_thread_info(True, 1)
        if thread_info is None:
            return

    thread = thread_info.thread
    if thread is None:
        return

    info = thread_info.additional_info
    if info is None:
        return

    # print('LINE %s %s %s %s' % (frame.f_lineno, code.co_name, code.co_filename, info.pydev_step_cmd))

    if info.is_tracing:
        return

    try:
        info.is_tracing = True

        py_db = GlobalDebuggerHolder.global_dbg

        if py_db is None:
            return monitoring.DISABLE

        if py_db.pydb_disposed:
            return monitoring.DISABLE

        stop_frame = info.pydev_step_stop
        step_cmd = info.pydev_step_cmd

        filename = _get_abs_path_real_path_and_base_from_frame(frame)[1]
        breakpoints_for_file = py_db.breakpoints.get(filename)

        frame_cache_key = _make_frame_cache_key(code)
        line_cache_key = (frame_cache_key, line_number)

        try:
            flag = False
            breakpoint = None
            stop = False
            exist_result = False
            bp_type = None
            args = (py_db, filename, info, thread)
            smart_stop_frame = info.pydev_smart_step_context.smart_step_stop
            context_start_line = info.pydev_smart_step_context.start_line
            context_end_line = info.pydev_smart_step_context.end_line
            is_within_context = (context_start_line <= line_number
                                 <= context_end_line)

            if info.pydev_state != STATE_SUSPEND and breakpoints_for_file is not None and line_number in breakpoints_for_file:
                breakpoint = breakpoints_for_file[line_number]
                new_frame = frame
                stop = True
                if step_cmd == CMD_STEP_OVER:
                    if stop_frame is frame:
                        stop = False
                elif step_cmd == CMD_SMART_STEP_INTO and (
                        frame.f_back is smart_stop_frame and is_within_context):
                    stop = False
            elif py_db.plugin is not None and py_db.has_plugin_line_breaks:
                result = py_db.plugin.get_breakpoint(py_db, frame, 'line', args)
                if result:
                    exist_result = True
                    flag, breakpoint, new_frame, bp_type = result

            if breakpoint:
                if stop or exist_result:
                    eval_result = False
                    if breakpoint.has_condition:
                        eval_result = handle_breakpoint_condition(py_db, info, breakpoint, new_frame)

                    if breakpoint.expression is not None:
                        handle_breakpoint_expression(breakpoint, info, new_frame)
                        if breakpoint.is_logpoint and info.pydev_message is not None and len(
                                info.pydev_message) > 0:
                            cmd = py_db.cmd_factory.make_io_message(info.pydev_message + os.linesep, '1')
                            py_db.writer.add_command(cmd)

                    if breakpoint.has_condition and not eval_result:
                        return
            else:
                if step_cmd != -1:
                    if (py_db.is_filter_enabled
                            and py_db.is_ignored_by_filters(filename)):
                        # ignore files matching stepping filters
                        return monitoring.DISABLE
                    if (py_db.is_filter_libraries
                            and not py_db.in_project_scope(filename)):
                        # ignore library files while stepping
                        return monitoring.DISABLE

            if py_db.show_return_values or py_db.remove_return_values_flag:
                manage_return_values(py_db, frame, 'line', None)

            if stop:
                py_db.set_suspend(
                    thread,
                    CMD_SET_BREAK,
                    suspend_other_threads=breakpoint
                                          and breakpoint.suspend_policy == "ALL",
                )
            elif flag and py_db.plugin is not None:
                result = py_db.plugin.suspend(py_db, thread, frame, bp_type)
                if result:
                    frame = result

            # if thread has a suspend flag, we suspend with a busy wait
            if info.pydev_state == STATE_SUSPEND:
                py_db.do_wait_suspend(thread, frame, 'line', None)
                return
            elif not breakpoint:
                # No stop from anyone and no breakpoint found in line (cache that).
                global_cache_frame_skips[line_cache_key] = 0
        except KeyboardInterrupt:
            _clear_run_state(info)
            raise
        except:
            traceback.print_exc()
            raise

        # Step handling. We stop when we hit the right frame.
        try:
            plugin_stop = False
            args = (py_db, filename, info, thread)
            stop_info = {}
            stop = False

            if step_cmd == CMD_SMART_STEP_INTO:
                if smart_stop_frame is frame and not is_within_context:
                    # We don't stop on jumps in multiline statements, which
                    # the Python interpreter does in some cases, if we they
                    # happen in smart step into context.
                    info.pydev_func_name = '.invalid.'  # Must match the type in cython
                    stop = True  # act as if we did a step into

                curr_func_name = frame.f_code.co_name

                if curr_func_name in ('?', '<module>') or curr_func_name is None:
                    curr_func_name = ''

                if smart_stop_frame and smart_stop_frame is frame.f_back:
                    if curr_func_name == info.pydev_func_name and not IS_CPYTHON:
                        stop = True
                    else:
                        try:
                            if curr_func_name != info.pydev_func_name and frame.f_back:
                                # try to find function call name using bytecode analysis
                                curr_func_name = find_last_call_name(frame.f_back)
                            if curr_func_name == info.pydev_func_name:
                                stop = find_last_func_call_order(frame.f_back, context_start_line) \
                                       == info.pydev_smart_step_context.call_order
                        except:
                            pydev_log.debug("Exception while handling smart step into "
                                            "in frame tracer, step into will be "
                                            "performed instead.")
                            info.pydev_smart_step_context.reset()
                            stop = True  # act as if we did a step into

            elif step_cmd == CMD_STEP_INTO:
                stop = True
                if py_db.plugin is not None:
                    result = py_db.plugin.cmd_step_into(py_db, frame, 'line', args, stop_info, stop)
                    if result:
                        stop, plugin_stop = result

            elif step_cmd == CMD_STEP_INTO_MY_CODE:
                stop = py_db.in_project_scope(frame.f_code.co_filename)

            elif step_cmd in (CMD_STEP_OVER, CMD_STEP_INTO_COROUTINE):
                stop = stop_frame is frame
                if stop:
                    # The only case we shouldn't stop on a line, is when
                    # we are traversing though asynchronous framework machinery
                    if step_cmd == CMD_STEP_INTO_COROUTINE:
                        stop = py_db.in_project_scope(frame.f_code.co_filename)

                if step_cmd == CMD_STEP_OVER and py_db.plugin is not None:
                    result = py_db.plugin.cmd_step_over(py_db, frame, 'line', args, stop_info, stop)
                    if result:
                        stop, plugin_stop = result
            else:
                stop = False

            if plugin_stop:
                py_db.plugin.stop(py_db, frame, 'line', args, stop_info, None, step_cmd)
            elif stop:
                py_db.set_suspend(thread, step_cmd)
                py_db.do_wait_suspend(thread, frame, 'line', None)

        except KeyboardInterrupt:
            _clear_run_state(info)
            raise
        except:
            traceback.print_exc()
            info.pydev_step_cmd = -1
            raise

        if py_db.quitting:
            raise KeyboardInterrupt()
    finally:
        info.is_tracing = False


def py_raise_callback(code, instruction_offset, exception):
    # print('PY_RAISE %s %s %s' % (code.co_name, code.co_filename, exception))

    exc_info = (type(exception), exception, exception.__traceback__)

    try:
        py_db = GlobalDebuggerHolder.global_dbg
    except AttributeError:
        py_db = None

    if py_db is None:
        return

    try:
        try:
            thread_info = _thread_local_info.thread_info
        except:
            thread_info = _get_thread_info(True, 1)
            if thread_info is None:
                return

        thread = thread_info.thread
        info = thread_info.additional_info
        if info is None:
            return

        frame = _getframe(1)
        if frame is _get_top_level_frame():
            _stop_on_unhandled_exception(exc_info, py_db, thread)
            return

        has_exception_breakpoints = (py_db.break_on_caught_exceptions
                                     or py_db.has_plugin_exception_breaks
                                     or py_db.stop_on_failed_tests)
        if has_exception_breakpoints:
            args = (
                py_db,
                _get_abs_path_real_path_and_base_from_frame(frame)[1],
                info, thread,
                global_cache_skips,
                global_cache_frame_skips
            )
            should_stop, frame = should_stop_on_exception(
                args, frame, 'exception', exc_info)
            if should_stop:
                handle_exception(args, frame, 'exception', exc_info)
    except KeyboardInterrupt:
        _clear_run_state(info)
        raise
    except:
        traceback.print_exc()
        info.pydev_step_cmd = -1
        raise

    if py_db.quitting:
        raise KeyboardInterrupt()


def py_return_callback(code, instruction_offset, retval):
    # print('PY_RETURN %s %s %s' % (code, code.co_name, code.co_filename))
    try:
        py_db = GlobalDebuggerHolder.global_dbg
    except AttributeError:
        py_db = None

    if py_db is None:
        return monitoring.DISABLE

    frame = _getframe(1)
    try:
        thread_info = _thread_local_info.thread_info
    except:
        thread_info = _get_thread_info(True, 1)
        if thread_info is None:
            return

    thread = thread_info.thread
    if thread is None:
        return

    info = thread_info.additional_info
    if info is None:
        return

    stop_frame = info.pydev_step_stop
    filename = _get_abs_path_real_path_and_base_from_frame(frame)[1]
    plugin_stop = False
    args = (py_db, filename, info, thread)
    stop_info = {}
    stop = False
    smart_stop_frame = info.pydev_smart_step_context.smart_step_stop

    try:
        if py_db.show_return_values or py_db.remove_return_values_flag:
            manage_return_values(py_db, frame, 'return', retval)

        step_cmd = info.pydev_step_cmd

        if step_cmd == CMD_SMART_STEP_INTO and smart_stop_frame is frame:
            stop = True

        elif step_cmd == CMD_STEP_INTO:
            stop = True
            if py_db.plugin is not None:
                result = py_db.plugin.cmd_step_into(py_db, frame, 'return', args, stop_info, True)
                if result:
                    stop, plugin_stop = result

        elif step_cmd in (CMD_STEP_OVER, CMD_STEP_INTO_COROUTINE):
            stop = stop_frame is frame
            if stop:
                stop = frame.f_back and py_db.in_project_scope(frame.f_back.f_code.co_filename)
                if not stop:
                    if frame.f_back:
                        back = frame.f_back
                        info.pydev_step_stop = back
                        back_code = back.f_code
                        if not py_db.in_project_scope(back_code.co_filename):
                            stop = not step_cmd == CMD_STEP_INTO_COROUTINE

                    else:
                        # if there's no back frame, we just stop as soon as possible
                        info.pydev_step_cmd = CMD_STEP_INTO
                        info.pydev_step_stop = None

            if step_cmd == CMD_STEP_OVER and py_db.plugin is not None:
                result = py_db.plugin.cmd_step_over(py_db, frame, 'return', args, stop_info, stop)
                if result:
                    stop, plugin_stop = result

        elif step_cmd == CMD_STEP_RETURN:
            stop = stop_frame is frame

        if hasattr(frame, "f_back"):
            f_back = frame.f_back
            f_code = getattr(f_back, 'f_code', None)
            if f_code is not None:
                back_filename = f_code.co_filename
                base_back_filename = os.path.basename(back_filename)
                file_type = get_file_type(base_back_filename)
                if stop != (step_cmd == -1):
                    if file_type == PYDEV_FILE:
                        stop = False
                    elif not stop and step_cmd == -1:
                        # Check does f_back have breakpoint and should enable line events for f_back
                        breakpoints_for_back_file = py_db.breakpoints.get(back_filename)
                        if breakpoints_for_back_file is not None:
                            for breakpoint in breakpoints_for_back_file.values():
                                if breakpoint.func_name == f_code.co_name:
                                    if _should_enable_line_events_for_code(f_back,
                                                                           f_code,
                                                                           back_filename,
                                                                           info):
                                        _enable_line_tracing(f_code)
                                        _enable_return_tracing(f_code)
                                    break
                if stop:
                    _enable_line_tracing(f_code)
                    _enable_return_tracing(f_code)

        if plugin_stop:
            py_db.plugin.stop(py_db, frame, 'return', args, stop_info, None, step_cmd)
        elif stop:
            back = frame.f_back
            if back is not None:
                _, base_back_filename, base = get_abs_path_real_path_and_base_from_frame(back)
                if (base, back.f_code.co_name) in (DEBUG_START, DEBUG_START_PY3K):
                    back = None

            if back is not None:
                # if we're in a return, we want it to appear to the user in the previous frame!
                py_db.set_suspend(thread, step_cmd)
                py_db.do_wait_suspend(thread, back, 'return', retval)
    except KeyboardInterrupt:
        _clear_run_state(info)
        raise
    except:
        traceback.print_exc()
        info.pydev_step_cmd = -1
        raise

    if py_db.quitting:
        raise KeyboardInterrupt()

def disable_pep669_monitoring(all_threads=False):
    if all_threads:
        monitoring.set_events(monitoring.DEBUGGER_ID, 0)
        monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.PY_START, None)
        monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.LINE, None)
        monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.PY_RETURN, None)
        monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.RAISE, None)
        monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.CALL, None)
        monitoring.free_tool_id(monitoring.DEBUGGER_ID)
    else:
        try:
            thread_info = _thread_local_info.thread_info
        except:
            thread_info = _get_thread_info(False, 1)
            if thread_info is None:
                return
        # print('stop monitoring, thread=', thread_info.thread)
        thread_info.trace = False


#  Copyright 2000-2023 JetBrains s.r.o. and contributors. Use of this source code is governed by the Apache 2.0 license.
import polars as pl

TABLE_TYPE_NEXT_VALUE_SEPARATOR = '__pydev_table_column_type_val__'
MAX_COLWIDTH = 100000
pl_version_major, pl_version_minor, _ = pl.__version__.split(".")
pl_version_major, pl_version_minor = int(pl_version_major), int(pl_version_minor)
COUNT_COL_NAME = "counts" if pl_version_major == 0 and pl_version_minor < 20 else "count"
CSV_FORMAT_SEPARATOR = '~'
DASH_SYMBOL = '\u2014'


class ColumnVisualisationType:
    HISTOGRAM = "histogram"
    UNIQUE = "unique"
    PERCENTAGE = "percentage"


class ColumnVisualisationUtils:
    NUM_BINS = 20
    MAX_UNIQUE_VALUES = 3
    MAX_UNIQUE_VALUES_TO_SHOW_IN_VIS = 50
    MAX_VALUES_LENGTH = 100

    TABLE_OCCURRENCES_COUNT_NEXT_COLUMN_SEPARATOR = '__pydev_table_occurrences_count_next_column__'
    TABLE_OCCURRENCES_COUNT_NEXT_VALUE_SEPARATOR = '__pydev_table_occurrences_count_next_value__'
    TABLE_OCCURRENCES_COUNT_DICT_SEPARATOR = '__pydev_table_occurrences_count_dict__'
    TABLE_OCCURRENCES_COUNT_OTHER = '__pydev_table_other__'


def get_type(table):
    # type: (Union[pl.Series, pl.DataFrame]) -> str
    return str(type(table))


def get_shape(table):
    # type: (Union[pl.Series, pl.DataFrame]) -> str
    return str(table.shape)


def get_head(table):
    # type: (Union[pl.Series, pl.DataFrame]) -> str
    with __create_config():
        return table.head(1)._repr_html_()


def get_column_types(table):
    # type: (Union[pl.Series, pl.DataFrame]) -> str
    if type(table) == pl.Series:
        return str(table.dtype)
    else:
        return TABLE_TYPE_NEXT_VALUE_SEPARATOR.join([str(t) for t in table.dtypes])


# used by pydevd
def get_data(table, use_csv_serialization, start_index=None, end_index=None, format=None):
    # type: (Union[pl.Series, pl.DataFrame], int, int) -> str
    with __create_config(format):
        if use_csv_serialization:
            float_precision = __get_float_precision(format)
            return __write_to_csv(__get_df_slice(table, start_index, end_index), float_precision=float_precision)
        return table[start_index:end_index]._repr_html_()


# used by DSTableCommands
def display_data_html(table, start_index, end_index):
    # type: (Union[pl.Series, pl.DataFrame], int, int) -> None
    with __create_config():
        print(table[start_index:end_index]._repr_html_())


def display_data_csv(table, start_index, end_index):
    # type: (Union[pl.Series, pl.DataFrame], int, int) -> None
    with __create_config():
        print(repr(__write_to_csv(__get_df_slice(table, start_index, end_index))))


def get_column_descriptions(table):
    # type: (Union[pl.Series, pl.DataFrame]) -> str
    described_results = __get_describe(table)

    if described_results is not None:
        return get_data(described_results, None, None)
    else:
        return ""


def get_value_occurrences_count(table):
    # type: (Union[pl.Series, pl.DataFrame]) -> str
    bin_counts = []
    if type(table) == pl.DataFrame:
        for col in table.columns:
            bin_counts.append(__analyze_column(col, table[col]))
    elif type(table) == pl.Series:
        bin_counts.append(__analyze_column(table.name, table))

    return ColumnVisualisationUtils.TABLE_OCCURRENCES_COUNT_NEXT_COLUMN_SEPARATOR.join(bin_counts)


def __analyze_column(col_name, column):
    col_type = column.dtype
    res = []
    column_visualisation_type = None

    if __is_boolean(column, col_type):
        column_visualisation_type, res = __analyze_boolean_column(column, col_name)

    elif __is_numeric(column, col_type):
        column_visualisation_type, res = __analyze_numeric_column(column, col_name)

    else:
        column_visualisation_type, res = __analyze_categorical_column(column, col_name)

    if column_visualisation_type != ColumnVisualisationType.UNIQUE:
        counts = ColumnVisualisationUtils.TABLE_OCCURRENCES_COUNT_NEXT_VALUE_SEPARATOR.join(res)
    else:
        counts = res

    return str({column_visualisation_type:counts})


def __is_boolean(column, col_type):
    # (pl.Series, pl.DataType) -> bool
    return col_type == pl.Boolean and not column.is_null().any()


def __is_numeric(column, col_type):
    # (pl.Series, pl.DataType) -> bool
    return __is_series_numeric(column) and not column.is_null().all()


def __analyze_boolean_column(column, col_name):
    counts = column.value_counts().sort(by=col_name).to_dict()
    return ColumnVisualisationType.HISTOGRAM, __add_custom_key_value_separator(zip(counts[col_name], counts[COUNT_COL_NAME]))


def __analyze_categorical_column(column, col_name):
    all_values = column.shape[0]
    if column.is_null().all():
        value_counts = pl.DataFrame({col_name: "None", COUNT_COL_NAME: all_values})
    else:
        value_counts = column.value_counts()

    # Sort in descending order to get values with max percent
    value_counts = value_counts.sort(COUNT_COL_NAME).reverse()

    if len(value_counts) <= ColumnVisualisationUtils.MAX_UNIQUE_VALUES or len(value_counts) / all_values * 100 <= ColumnVisualisationUtils.MAX_UNIQUE_VALUES_TO_SHOW_IN_VIS:
        column_visualisation_type = ColumnVisualisationType.PERCENTAGE

        # If column contains <= 3 unique values no `Other` category is shown, but all of these values and their percentages
        num_unique_values_to_show_in_vis = ColumnVisualisationUtils.MAX_UNIQUE_VALUES - (0 if len(value_counts) == 3 else 1)
        counts = value_counts[:num_unique_values_to_show_in_vis]
        counts = counts.with_columns(((pl.col(COUNT_COL_NAME) / all_values * 100).round(1)).alias(COUNT_COL_NAME))
        top_values = {}
        for label, count in zip(counts[col_name], counts[COUNT_COL_NAME]):
            # we should process separately a case with dtype == pl.List
            if type(label) == pl.Series or column.dtype == pl.List:
                label_values = label.to_list()
                label_values_in_str = str(label_values)
                top_values[label_values_in_str[:ColumnVisualisationUtils.MAX_VALUES_LENGTH]] = count
            else:
                label_in_str = str(label)
                top_values[label_in_str[:ColumnVisualisationUtils.MAX_VALUES_LENGTH]] = count
        if len(value_counts) == ColumnVisualisationUtils.MAX_UNIQUE_VALUES:
            top_values[ColumnVisualisationUtils.TABLE_OCCURRENCES_COUNT_OTHER] = -1
        else:
            others_count = value_counts[num_unique_values_to_show_in_vis:][COUNT_COL_NAME].sum()
            top_values[ColumnVisualisationUtils.TABLE_OCCURRENCES_COUNT_OTHER] = round(others_count / all_values * 100, 1)
        res = __add_custom_key_value_separator(top_values.items())

    else:
        column_visualisation_type = ColumnVisualisationType.UNIQUE
        res = len(value_counts)
    return column_visualisation_type, res


def __analyze_numeric_column(column, col_name):
    # handle np.NaN values, because they are not dropped with drop_nulls() the way they are
    column = column.fill_null(strategy="min")
    if column.shape[0] <= ColumnVisualisationUtils.NUM_BINS:
        raw_counts = column.value_counts().sort(by=col_name).to_dict(as_series=False)
        res = __add_custom_key_value_separator(zip(raw_counts[col_name], raw_counts[COUNT_COL_NAME]))
    else:
        import numpy as np

        def format_function(x):
            if x == int(x):
                return int(x)
            else:
                return round(x, 3)

        counts, bin_edges = np.histogram(column, bins=ColumnVisualisationUtils.NUM_BINS)
        # so the long dash will be correctly viewed both on Mac and Windows
        bin_labels = ['{} {} {}'.format(format_function(bin_edges[i]), DASH_SYMBOL, format_function(bin_edges[i + 1])) for i in range(ColumnVisualisationUtils.NUM_BINS)]
        res = __add_custom_key_value_separator(zip(bin_labels, counts))

    return ColumnVisualisationType.HISTOGRAM, res


def __get_df_slice(table, start_index, end_index):
    # type: (Union[pl.Series, pl.DataFrame], int, int) -> pl.DataFrame
    if type(table) == pl.Series:
        return table[start_index:end_index].to_frame()
    return table[start_index:end_index]


def __write_to_csv(table, null_value="null", float_precision=None):
    def serialize_nested(value, null_value="null", float_precision=None):
        if value is None:
            return null_value
        elif isinstance(value, float) and float_precision is not None:
            return "{:.{}f}".format(value, float_precision)
        elif isinstance(value, dict):
            return "{" + ", ".join("{}: {}".format(k, serialize_nested(v, null_value, float_precision)) for k, v in value.items()) + "}"
        elif isinstance(value, list):
            return "[" + ", ".join(serialize_nested(v, null_value, float_precision) for v in value) + "]"
        else:
            return str(value)

    lines = []
    lines.append(CSV_FORMAT_SEPARATOR.join(table.columns))
    for row in table.rows():
        line = []
        for value in row:
            line.append(serialize_nested(value, null_value, float_precision))
        lines.append(CSV_FORMAT_SEPARATOR.join(line))
    return "\n".join(lines)


def __create_config(format=None):
    # type: (Union[str, None]) -> pl.Config
    cfg = pl.Config()
    cfg.set_tbl_cols(-1)  # Unlimited
    cfg.set_tbl_rows(-1)  # Unlimited
    cfg.set_fmt_str_lengths(MAX_COLWIDTH)  # No option to set unlimited, so it's 100_000
    float_precision = __get_float_precision(format)
    if float_precision is not None and hasattr(cfg, 'set_float_precision'):
        cfg.set_float_precision(float_precision)
    return cfg


def __add_custom_key_value_separator(pairs_list):
    return [str(label) + ColumnVisualisationUtils.TABLE_OCCURRENCES_COUNT_DICT_SEPARATOR + str(count) for label, count in pairs_list]


def __get_describe(table):
    # type: (Union[pl.Series, pl.DataFrame]) -> Union[pl.DataFrame, None]
    try:
        if type(table) == pl.DataFrame and 'describe' in table.columns:
            import random
            random_suffix = ''.join([chr(random.randint(97, 122)) for _ in range(5)])
            described_df = table\
                .rename({'describe': 'describe_original_' + random_suffix})\
                .describe(percentiles=(0.05, 0.25, 0.5, 0.75, 0.95))
        else:
            described_df = table.describe(percentiles=(0.05, 0.25, 0.5, 0.75, 0.95))
        return described_df
    # If DataFrame/Series have unsupported type for describe
    # then Polars will raise TypeError exception. We should catch them.
    except Exception as e:
        return


def __get_float_precision(format):
    # type: (Union[str, None]) -> Union[int, None]
    if isinstance(format, str):
        if format.startswith("%") and format.endswith("f"):
            start = format.find('%.') + 2
            end = format.find('f')

            if start < end:
                try:
                    precision = int(format[start:end])
                    return precision
                except:
                    pass

    return None


def __is_series_numeric(column):
    """
    Determines if the given column is numeric based on the version of the polars
    library being used.

    For polars major version 0, the method checks if the column is numeric using
    the `is_numeric` method directly on the column. For later versions, it checks
    the `is_numeric` method on the column's data type.
    """
    if pl_version_major == 0:
        return column.is_numeric()
    else:
        return column.dtype.is_numeric()



#  Copyright 2000-2023 JetBrains s.r.o. and contributors. Use of this source code is governed by the Apache 2.0 license.
import os
import sys
import subprocess

IS_WINDOWS = sys.platform == 'win32'
python_envs_path = os.getenv('ENVS_DIRECTORY')
if IS_WINDOWS:
    python_installations = [
        r'%s\py27_64\Scripts\python.exe' % python_envs_path,
        r'%s\py36_64\Scripts\python.exe' % python_envs_path,
        r'%s\py37_64\Scripts\python.exe' % python_envs_path,
        r'%s\py38_64\Scripts\python.exe' % python_envs_path,
        r'%s\py39_64\Scripts\python.exe' % python_envs_path,
        r'%s\py310_64\Scripts\python.exe' % python_envs_path,
        r'%s\py311_64\Scripts\python.exe' % python_envs_path,
    ]
else:
    python_installations = [
        r'%s/py27_64/bin/python' % python_envs_path,
        r'%s/py36_64/bin/python' % python_envs_path,
        r'%s/py37_64/bin/python' % python_envs_path,
        r'%s/py38_64/bin/python' % python_envs_path,
        r'%s/py39_64/bin/python' % python_envs_path,
        r'%s/py310_64/bin/python' % python_envs_path,
        r'%s/py311_64/bin/python' % python_envs_path,
    ]
GENERATOR_FILE_NAME = "generate_import_error.py"
All_ERRORS_LINES_FILE = os.path.join(os.path.dirname(__file__), 'error_lines.txt')
RESULT_FILE = os.path.join(os.path.dirname(__file__), 'result.txt')
RUN_GEN_STRING = 'Run generation with python: %s'
GEN_FINISHED_STRING = 'Generation finished'


def compute_result_output():
    with open(All_ERRORS_LINES_FILE, 'r') as res_file:
        lines = res_file.readlines()
    out = set()
    for line in lines:
        out.add(line)

    # if we have results from another OS
    try:
        with open(RESULT_FILE, 'r') as res_file:
            lines = res_file.readlines()
        for line in lines:
            out.add(line)
    except:
        pass

    with open(RESULT_FILE, 'w') as res_file:
        for line in out:
            res_file.write(line)
    os.remove(os.path.join(All_ERRORS_LINES_FILE))


if __name__ == '__main__':
    print("Pythons: %s", python_installations)
    for python in python_installations:
        print(RUN_GEN_STRING % python)
        p = subprocess.Popen([python, os.path.join(os.path.dirname(__file__), GENERATOR_FILE_NAME)])
        p.wait()
        print(GEN_FINISHED_STRING)

    compute_result_output()



"""Bytecode analyzing utils."""

from _pydevd_bundle.smart_step_into import get_stepping_variants
from _pydevd_bundle.pydevd_constants import IS_PY311_OR_GREATER, IS_PY39_OR_GREATER
import dis


def find_last_func_call_order(frame, start_line):
    """Find the call order of the last function call between ``start_line``
    and last executed instruction.

    :param frame: A frame inside which we are looking for the function call.
    :type frame: :py:class:`types.FrameType`
    :param start_line:
    :return: call order or -1 if we fail to find the call order for some
      reason.
    :rtype: int
    """
    code = frame.f_code
    lasti = frame.f_lasti
    cache = {}
    call_order = -1

    def process_instructions(order, start_offset=None, end_offset=None, find_reraise=False):
        """Processes instructions within the given bytecode offsets."""
        stepping_generator = get_stepping_variants(code)
        for inst in stepping_generator:
            if end_offset is not None and inst.offset > end_offset:
                if not find_reraise:
                    break
                else:
                    for forward in get_stepping_variants(code):
                        if forward.lineno > frame.f_lineno:
                            break
                        cache_for_reraise = cache.copy()
                        name = forward.argval
                        call_order_reraise = cache_for_reraise.setdefault(name, -1)
                        call_order_reraise += 1
                        cache_for_reraise[name] = call_order_reraise
                        if name == "RERAISE":
                            order = call_order_reraise
                            break
                    break

            if start_offset is not None and inst.offset < start_offset:
                continue
            if start_line <= inst.lineno <= frame.f_lineno:
                name = inst.argval
                order = cache.setdefault(name, -1)
                order += 1
                cache[name] = order
        return order

    if IS_PY311_OR_GREATER:
        exception_table = dis._parse_exception_table(code)
        if exception_table:
            for i, entry in enumerate(exception_table):
                if entry.start < lasti < entry.end:
                    target_entry = None
                    for next_entry in exception_table[i + 1:]:
                        if next_entry.start == entry.target:
                            target_entry = next_entry
                            break
                    if target_entry is not None:
                        call_order = process_instructions(call_order, target_entry.start, target_entry.end)
                        break
            else:
                call_order = process_instructions(call_order, None, lasti)
        else:
            call_order = process_instructions(call_order, None, lasti)
    elif IS_PY39_OR_GREATER:
        call_order = process_instructions(call_order, None, lasti, find_reraise=True)
    else:
        call_order = process_instructions(call_order, None, lasti)

    return call_order


def find_last_call_name(frame):
    """Find the name of the last call made in the frame.

    :param frame: A frame inside which we are looking the last call.
    :type frame: :py:class:`types.FrameType`
    :return: The name of a function or method that has been called last.
    :rtype: str
    """
    last_call_name = None
    code = frame.f_code
    last_instruction = frame.f_lasti
    for inst in get_stepping_variants(code):
        if inst.offset > last_instruction:
            break
        last_call_name = inst.argval

    return last_call_name



# Defines which version of the PyDBAdditionalThreadInfo we'll use.

import os
use_cython = os.getenv('PYDEVD_USE_CYTHON', None)

if use_cython == 'YES':
    # We must import the cython version if forcing cython
    from _pydevd_bundle.pydevd_cython_wrapper import PyDBAdditionalThreadInfo, set_additional_thread_info, _set_additional_thread_info_lock  # @UnusedImport

elif use_cython == 'NO':
    # Use the regular version if not forcing cython
    from _pydevd_bundle.pydevd_additional_thread_info_regular import PyDBAdditionalThreadInfo, set_additional_thread_info, _set_additional_thread_info_lock  # @UnusedImport @Reimport

elif use_cython is None:
    # Regular: use fallback if not found (message is already given elsewhere).
    try:
        from _pydevd_bundle.pydevd_cython_wrapper import PyDBAdditionalThreadInfo, set_additional_thread_info, _set_additional_thread_info_lock
    except ImportError:
        from _pydevd_bundle.pydevd_additional_thread_info_regular import PyDBAdditionalThreadInfo, set_additional_thread_info, _set_additional_thread_info_lock  # @UnusedImport
else:
    raise RuntimeError('Unexpected value for PYDEVD_USE_CYTHON: %s (accepted: YES, NO)' % (use_cython,))





#  Copyright 2000-2024 JetBrains s.r.o. and contributors. Use of this source code is governed by the Apache 2.0 license.
def get_apply():
    try:
        from pydevd_nest_asyncio import apply
        return apply
    except:
        return None


def get_eval_async_expression_in_context():
    try:
        from pydevd_asyncio_utils import eval_async_expression_in_context
        return eval_async_expression_in_context
    except:
        return None


def get_eval_async_expression():
    try:
        from pydevd_asyncio_utils import eval_async_expression
        return eval_async_expression
    except:
        return None


def get_exec_async_code():
    try:
        from pydevd_asyncio_utils import exec_async_code
        return exec_async_code
    except:
        return None


def get_asyncio_command_compiler():
    try:
        from pydevd_asyncio_utils import asyncio_command_compiler
        return asyncio_command_compiler
    except:
        return None



import sys

# This version number is always available
from _pydevd_bundle.pydevd_additional_thread_info_regular import version as regular_version

try:
    try:
        from _pydevd_bundle_ext import pydevd_cython as mod

    except ImportError:
        from _pydevd_bundle import pydevd_cython as mod

except ImportError:
    import struct

    try:
        is_python_64bit = (struct.calcsize('P') == 8)
    except:
        # In Jython this call fails, but this is Ok, we don't support Jython for speedups anyways.
        raise ImportError
    plat = '32'
    if is_python_64bit:
        plat = '64'

    # We also accept things as:
    #
    # _pydevd_bundle.pydevd_cython_win32_27_32
    # _pydevd_bundle.pydevd_cython_win32_34_64
    #
    # to have multiple pre-compiled pyds distributed along the IDE
    # (generated by build_tools/build_binaries_windows.py).

    mod_name = 'pydevd_cython_%s_%s%s_%s' % (sys.platform, sys.version_info[0], sys.version_info[1], plat)
    check_name = '_pydevd_bundle.%s' % (mod_name,)
    mod = getattr(__import__(check_name), mod_name)

# Regardless of how it was found, make sure it's later available as the
# initial name so that the expected types from Cython in frame eval
# are valid.
sys.modules['_pydevd_bundle.pydevd_cython'] = mod

# This version number from the already compiled Cython extension
version = getattr(mod, 'version', 0)

if version != regular_version:
    exc = ImportError()
    exc.version_mismatch = True
    raise exc

trace_dispatch = mod.trace_dispatch

PyDBAdditionalThreadInfo = mod.PyDBAdditionalThreadInfo

set_additional_thread_info = mod.set_additional_thread_info

global_cache_skips = mod.global_cache_skips

global_cache_frame_skips = mod.global_cache_frame_skips

_set_additional_thread_info_lock = mod._set_additional_thread_info_lock

fix_top_level_trace_and_get_trace_func = mod.fix_top_level_trace_and_get_trace_func

# Common frame functions to be used in PEP 669 debugger callbacks.
handle_breakpoint_condition = mod.handle_breakpoint_condition
handle_breakpoint_expression = mod.handle_breakpoint_expression
DEBUG_START = mod.DEBUG_START
DEBUG_START_PY3K = mod.DEBUG_START_PY3K
should_stop_on_exception = mod.should_stop_on_exception
handle_exception = mod.handle_exception
manage_return_values = mod.manage_return_values



#  Copyright 2000-2022 JetBrains s.r.o. and contributors. Use of this source code is governed by the Apache 2.0 license.
import threading

from _pydevd_bundle.pydevd_constants import RETURN_VALUES_DICT, dict_iter_items, \
    GET_FRAME_NORMAL_GROUP, GET_FRAME_SPECIAL_GROUP

HIDDEN_TYPES = ('function', 'type', 'classobj', 'module', 'typing')
DOUBLE_UNDERSCORE = '__'
DOUBLE_EX = '__exception__'
DUMMY_RET_VAL = '_dummy_ret_val'
DUMMY_IPYTHON_HIDDEN = '_dummy_ipython_val'
DUMMY_SPECIAL_VAR = '_dummy_special_var'
DO_NOT_PROCESS_VARS = (DUMMY_SPECIAL_VAR, DUMMY_RET_VAL, DUMMY_IPYTHON_HIDDEN)


class Handler(object):
    def __init__(self, fun):
        self.lst = []
        self.fun = fun

    _next_handler = None

    def set_next(self, handler):
        self._next_handler = handler
        return handler

    def handle(self, key, value, hidden_ns, evaluate_full_value, user_type_renderers=None):
        self.give_to_next(key, value, hidden_ns, evaluate_full_value, user_type_renderers)

    def regular_handle(self, key, value, hidden_ns, evaluate_full_value, is_dict_iter_items=False, user_type_renderers=None):
        if self.is_belong_to_group(key, value, hidden_ns):
            if is_dict_iter_items:
                for name, val in dict_iter_items(value):
                    self.lst.append(self.fun(name, val, hidden_ns, evaluate_full_value, user_type_renderers))
            else:
                self.lst.append(self.fun(key, value, hidden_ns, evaluate_full_value, user_type_renderers))
        else:
            self.give_to_next(key, value, hidden_ns, evaluate_full_value, user_type_renderers)

    def give_to_next(self, key, value, hidden_ns, evaluate_full_value, user_type_renderers):
        if self._next_handler:
            self._next_handler.handle(key, value, hidden_ns, evaluate_full_value, user_type_renderers)

    @staticmethod
    def is_belong_to_group(key, value, hidden_ns):
        return False

    def get_xml(self):
        xml = ''
        handler = self
        while handler is not None:
            for value in handler.lst:
                xml += value
            handler = handler._next_handler
        return xml

    def get_list(self):
        result = []
        handler = self
        while handler is not None:
            result += handler.lst
            handler = handler._next_handler
        return result

    def update_handlers(self):
        handler = self
        while handler is not None:
            handler.lst = []
            if hasattr(handler, 'added_var'):
                handler.added_var = False

            handler = handler._next_handler


class DunderVarsHandler(Handler):
    def handle(self, key, value, hidden_ns, evaluate_full_value, user_type_renderers=None):
        self.regular_handle(key, value, hidden_ns, evaluate_full_value, user_type_renderers)

    @staticmethod
    def is_belong_to_group(key, value, hidden_ns):
        return is_dunder_var(str(key))


class SpecialVarsHandler(Handler):
    def handle(self, key, value, hidden_ns, evaluate_full_value, user_type_renderers=None):
        self.regular_handle(key, value, hidden_ns, evaluate_full_value, user_type_renderers)

    @staticmethod
    def is_belong_to_group(key, value, hidden_ns):
        return is_hidden_var(value)


class IpythonVarsHandler(Handler):
    def handle(self, key, value, hidden_ns, evaluate_full_value, user_type_renderers=None):
        self.regular_handle(key, value, hidden_ns, evaluate_full_value, user_type_renderers)

    @staticmethod
    def is_belong_to_group(key, value, hidden_ns):
        return hidden_ns is not None and key in hidden_ns


class PytestVarsHandler(Handler):
    def handle(self, key, value, hidden_ns, evaluate_full_value, user_type_renderers=None):
        self.regular_handle(key, value, hidden_ns, evaluate_full_value, user_type_renderers)

    @staticmethod
    def is_belong_to_group(key, value, hidden_ns):
        return str(key).startswith('@')


class ReturnVarsHandler(Handler):
    def handle(self, key, value, hidden_ns, evaluate_full_value, user_type_renderers=None):
        self.regular_handle(key, value, hidden_ns, evaluate_full_value, True, user_type_renderers)

    @staticmethod
    def is_belong_to_group(key, value, hidden_ns):
        return key == RETURN_VALUES_DICT


class AnotherVarsHandler(Handler):
    def handle(self, key, value, hidden_ns, evaluate_full_value, user_type_renderers=None):
        if is_special_var(key, value):
            self.lst.append(self.fun(key, value, hidden_ns, evaluate_full_value, user_type_renderers))


class DefaultVarHandler(Handler):
    def handle(self, key, value, hidden_ns, evaluate_full_value, user_type_renderers=None):
        self.lst.append(self.fun(key, value, evaluate_full_value, user_type_renderers))


class DummyVarHandler(Handler):
    def __init__(self, fun, cls):
        super(DummyVarHandler, self).__init__(fun)
        self.cls = cls
        self.added_var = False

    def handle(self, key, value, hidden_ns, evaluate_full_value, user_type_renderers=None):
        if self.cls.is_belong_to_group(key, value, hidden_ns):
            if self.added_var:
                return
            self.lst.append(self.fun())
            self.added_var = True
        else:
            super(DummyVarHandler, self).handle(key, value, hidden_ns, evaluate_full_value, user_type_renderers)


def is_special_var(key, value):
    return is_hidden_var(value) or is_dunder_var(str(key))


def is_dunder_var(key_str):
    return key_str.startswith(DOUBLE_UNDERSCORE) and \
           key_str.endswith(DOUBLE_UNDERSCORE) and \
           key_str != DOUBLE_EX and \
           key_str != '__' and \
           key_str != '___'


def is_hidden_var(value):
    return get_type(value) in HIDDEN_TYPES


def get_type(o):
    try:
        type_object = o.__class__
    except:
        type_object = type(o)

    return type_object.__name__


XML_COMMUNICATION_VARS_HANDLER = 0
THRIFT_COMMUNICATION_VARS_HANDLER = 1


class VarsHandlerContainer:
    instance_xml_handler = None
    instance_thrift_handler = None


_vars_handler_initialization_lock = threading.RLock()


def _init_xml_communication_vars_handler(func, handler_type):
    with _vars_handler_initialization_lock:
        if VarsHandlerContainer.instance_xml_handler is None:
            VarsHandlerContainer.instance_xml_handler = \
                VarsHandler(func, handler_type)


def _init_thrift_communication_var_handler(func, handler_type):
    with _vars_handler_initialization_lock:
        if VarsHandlerContainer.instance_thrift_handler is None:
            VarsHandlerContainer.instance_thrift_handler = \
                VarsHandler(func, handler_type)


def get_vars_handler(func, handler_type, group_type):
    if handler_type == XML_COMMUNICATION_VARS_HANDLER:
        if VarsHandlerContainer.instance_xml_handler is None:
            _init_xml_communication_vars_handler(func, handler_type)
        return VarsHandlerContainer.instance_xml_handler.get_instance(group_type)
    elif handler_type == THRIFT_COMMUNICATION_VARS_HANDLER:
        if VarsHandlerContainer.instance_thrift_handler is None:
            _init_thrift_communication_var_handler(func, handler_type)
        return VarsHandlerContainer.instance_thrift_handler.get_instance(group_type)


class VarsHandler:
    def __init__(self, func, handler_type):
        self.func = func
        self._instance_normal = None
        self._instance_special = None
        self._instance_return = None
        self.handler_type = handler_type
        self._lock = threading.RLock()

    def get_instance(self, group_type):
        with self._lock:
            if group_type == GET_FRAME_NORMAL_GROUP:
                if self._instance_normal is None:
                    self._init_normal()
                instance = self._instance_normal
            elif group_type == GET_FRAME_SPECIAL_GROUP:
                if self._instance_special is None:
                    self._init_special()
                instance = self._instance_special
            else:
                if self._instance_return is None:
                    self._init_return()
                instance = self._instance_return

            instance.update_handlers()

            return instance

    def _init_normal(self):
        self._instance_normal = DummyVarHandler(
            lambda: self.func(DUMMY_RET_VAL, DUMMY_RET_VAL), ReturnVarsHandler
        )

        self._instance_normal.set_next(
            DummyVarHandler(
                lambda: self.func(DUMMY_IPYTHON_HIDDEN, DUMMY_IPYTHON_HIDDEN),
                IpythonVarsHandler
            )
        ).set_next(
            DummyVarHandler(
                lambda: self.func(DUMMY_SPECIAL_VAR, DUMMY_SPECIAL_VAR),
                SpecialVarsHandler
            )
        ).set_next(
            DummyVarHandler(
                lambda: self.func(DUMMY_SPECIAL_VAR, DUMMY_SPECIAL_VAR),
                PytestVarsHandler
            )
        ).set_next(
            DummyVarHandler(
                lambda: self.func(DUMMY_SPECIAL_VAR, DUMMY_SPECIAL_VAR),
                DunderVarsHandler
            )
        ).set_next(
            DefaultVarHandler(
                lambda key, var, eval, type_renderers: self.func(var,
                                                                 str(key),
                                                                 evaluate_full_value=eval,
                                                                 user_type_renderers=type_renderers)
            )
        )

    def _get_lambda_for_special(self, is_special_lambda=False):
        if self.handler_type == XML_COMMUNICATION_VARS_HANDLER:
            if is_special_lambda:
                additional_in_xml = ' isSpecialVal="True"'
            else:
                additional_in_xml = ' isIPythonHidden="True"'

            return lambda key, var, hidden_ns, eval_full, type_renderers: self.func(var,
                                                                                    str(key),
                                                                                    additional_in_xml=additional_in_xml,
                                                                                    evaluate_full_value=eval_full,
                                                                                    user_type_renderers=type_renderers)
        elif self.handler_type == THRIFT_COMMUNICATION_VARS_HANDLER:
            return lambda key, var, hidden_ns, eval_full, type_renderers: self.func(var,
                                                                                    str(key),
                                                                                    evaluate_full_value=eval_full,
                                                                                    user_type_renderers=type_renderers)
        else:
            raise ValueError('Handler type is incorrect')

    def _init_special(self):
        initial_lambda = self._get_lambda_for_special(is_special_lambda=False)
        special_lambda = self._get_lambda_for_special(is_special_lambda=True)

        self._instance_special = DunderVarsHandler(initial_lambda)
        self._instance_special.set_next(
            SpecialVarsHandler(special_lambda)
        ).set_next(
            IpythonVarsHandler(special_lambda)
        ).set_next(
            PytestVarsHandler(special_lambda)
        ).set_next(
            AnotherVarsHandler(special_lambda)
        )

    def _get_lambda_for_return(self):
        if self.handler_type == XML_COMMUNICATION_VARS_HANDLER:
            return lambda name, val, hidden_ns, eval_full, type_renderers: self.func(val,
                                                                                     name,
                                                                                     additional_in_xml=' isRetVal="True"',
                                                                                     user_type_renderers=type_renderers)
        elif self.handler_type == THRIFT_COMMUNICATION_VARS_HANDLER:
            return lambda name, val, hidden_ns, eval_full, type_renderers: self.func(val, name)
        else:
            return None

    def _init_return(self):
        self._instance_return = ReturnVarsHandler(self._get_lambda_for_return())




#=======================================================================================================================
# getopt code copied since gnu_getopt is not available on jython 2.1
#=======================================================================================================================
class GetoptError(Exception):
    opt = ''
    msg = ''
    def __init__(self, msg, opt=''):
        self.msg = msg
        self.opt = opt
        Exception.__init__(self, msg, opt)

    def __str__(self):
        return self.msg


def gnu_getopt(args, shortopts, longopts=[]):
    """getopt(args, options[, long_options]) -> opts, args

    This function works like getopt(), except that GNU style scanning
    mode is used by default. This means that option and non-option
    arguments may be intermixed. The getopt() function stops
    processing options as soon as a non-option argument is
    encountered.

    If the first character of the option string is `+', or if the
    environment variable POSIXLY_CORRECT is set, then option
    processing stops as soon as a non-option argument is encountered.
    """

    opts = []
    prog_args = []
    if type('') == type(longopts):
        longopts = [longopts]
    else:
        longopts = list(longopts)

    # Allow options after non-option arguments?
    all_options_first = False
    if shortopts.startswith('+'):
        shortopts = shortopts[1:]
        all_options_first = True

    while args:
        if args[0] == '--':
            prog_args += args[1:]
            break

        if args[0][:2] == '--':
            opts, args = do_longs(opts, args[0][2:], longopts, args[1:])
        elif args[0][:1] == '-':
            opts, args = do_shorts(opts, args[0][1:], shortopts, args[1:])
        else:
            if all_options_first:
                prog_args += args
                break
            else:
                prog_args.append(args[0])
                args = args[1:]

    return opts, prog_args

def do_longs(opts, opt, longopts, args):
    try:
        i = opt.index('=')
    except ValueError:
        optarg = None
    else:
        opt, optarg = opt[:i], opt[i + 1:]

    has_arg, opt = long_has_args(opt, longopts)
    if has_arg:
        if optarg is None:
            if not args:
                raise GetoptError('option --%s requires argument' % opt, opt)
            optarg, args = args[0], args[1:]
    elif optarg:
        raise GetoptError('option --%s must not have an argument' % opt, opt)
    opts.append(('--' + opt, optarg or ''))
    return opts, args

# Return:
#   has_arg?
#   full option name
def long_has_args(opt, longopts):
    possibilities = [o for o in longopts if o.startswith(opt)]
    if not possibilities:
        raise GetoptError('option --%s not recognized' % opt, opt)
    # Is there an exact match?
    if opt in possibilities:
        return False, opt
    elif opt + '=' in possibilities:
        return True, opt
    # No exact match, so better be unique.
    if len(possibilities) > 1:
        # XXX since possibilities contains all valid continuations, might be
        # nice to work them into the error msg
        raise GetoptError('option --%s not a unique prefix' % opt, opt)
    assert len(possibilities) == 1
    unique_match = possibilities[0]
    has_arg = unique_match.endswith('=')
    if has_arg:
        unique_match = unique_match[:-1]
    return has_arg, unique_match

def do_shorts(opts, optstring, shortopts, args):
    while optstring != '':
        opt, optstring = optstring[0], optstring[1:]
        if short_has_arg(opt, shortopts):
            if optstring == '':
                if not args:
                    raise GetoptError('option -%s requires argument' % opt,
                                      opt)
                optstring, args = args[0], args[1:]
            optarg, optstring = optstring, ''
        else:
            optarg = ''
        opts.append(('-' + opt, optarg))
    return opts, args

def short_has_arg(opt, shortopts):
    for i in range(len(shortopts)):
        if opt == shortopts[i] != ':':
            return shortopts.startswith(':', i + 1)
    raise GetoptError('option -%s not recognized' % opt, opt)


#=======================================================================================================================
# End getopt code
#=======================================================================================================================
